# Unit test predict ForecasterRecursiveMultiSeries
# ==============================================================================
import re
import pytest
import numpy as np
import pandas as pd
from sklearn.exceptions import NotFittedError
from sklearn.compose import ColumnTransformer
from sklearn.preprocessing import StandardScaler
from sklearn.preprocessing import MinMaxScaler
from sklearn.preprocessing import OneHotEncoder
from sklearn.preprocessing import OrdinalEncoder
from sklearn.preprocessing import FunctionTransformer
from sklearn.compose import make_column_transformer
from sklearn.compose import make_column_selector
from sklearn.pipeline import make_pipeline
from sklearn.linear_model import LinearRegression
from sklearn.ensemble import HistGradientBoostingRegressor
from lightgbm import LGBMRegressor

from skforecast.exceptions import IgnoredArgumentWarning
from skforecast.preprocessing import RollingFeatures
from skforecast.preprocessing import TimeSeriesDifferentiator
from ....recursive import ForecasterRecursiveMultiSeries

# Fixtures
from .fixtures_forecaster_recursive_multiseries import (
    series_wide_dt,
    series_dict_range,
    series_dict_dt,
    exog_wide_range,
    exog_long_dt,
    exog_pred_wide_range,
    exog_pred_long_dt,
    series_dict_nans_train,
    exog_dict_nans_train,
    exog_dict_nans_test,
    expected_df_to_long_format
)

series_2 = pd.DataFrame(
    {'1': pd.Series(np.arange(start=0, stop=50)), 
     '2': pd.Series(np.arange(start=50, stop=100))}
).to_dict(orient='series')


def test_predict_NotFittedError_when_fitted_is_False():
    """
    Test NotFittedError is raised when fitted is False.
    """
    forecaster = ForecasterRecursiveMultiSeries(LinearRegression(), lags=5)

    err_msg = re.escape(
        "This Forecaster instance is not fitted yet. Call `fit` with "
        "appropriate arguments before using predict."
    )
    with pytest.raises(NotFittedError, match = err_msg):
        forecaster.predict(steps=5)


def test_predict_IgnoredArgumentWarning_when_not_available_self_last_window_for_some_levels():
    """
    Test IgnoredArgumentWarning is raised when last_window is not available for 
    levels because it was not stored during fit.
    """
    forecaster = ForecasterRecursiveMultiSeries(LinearRegression(), lags=5)
    forecaster.fit(series=series_2, store_last_window=['1'])

    warn_msg = re.escape(
        "Levels {'2'} are excluded from prediction "
        "since they were not stored in `last_window_` attribute "
        "during training. If you don't want to retrain the "
        "Forecaster, provide `last_window` as argument."
    )
    with pytest.warns(IgnoredArgumentWarning, match = warn_msg):
        predictions = forecaster.predict(steps=5, levels=['1', '2'], last_window=None)

    expected = pd.DataFrame(
                   data    = np.array([50., 51., 52., 53., 54.]),
                   index   = pd.RangeIndex(start=50, stop=55, step=1),
                   columns = ['1']
               )

    expected = expected_df_to_long_format(expected)

    pd.testing.assert_frame_equal(predictions, expected)


@pytest.mark.parametrize("store_last_window",
                         [['1'], False],
                         ids=lambda slw: f"store_last_window: {slw}")
def test_predict_ValueError_when_not_available_self_last_window_for_levels(store_last_window):
    """
    Test ValueError is raised when last_window is not available for all 
    levels because it was not stored during fit.
    """
    forecaster = ForecasterRecursiveMultiSeries(LinearRegression(), lags=5)
    forecaster.fit(series=series_2, store_last_window=store_last_window)

    err_msg = re.escape(
        "No series to predict. None of the series {'2'} are present in "
        "`last_window_` attribute. Provide `last_window` as argument "
        "in predict method."
    )
    with pytest.raises(ValueError, match = err_msg):
        forecaster.predict(steps=5, levels=['2'], last_window=None)


def test_predict_IgnoredArgumentWarning_when_levels_is_list_and_different_last_index_in_self_last_window_DatetimeIndex():
    """
    Test IgnoredArgumentWarning is raised when levels is a list and have 
    different last index in last_window attribute using a DatetimeIndex.
    """
    series_3 = {
        '1': series_2['1'].copy(),
        '2': series_2['2'].iloc[:30].copy(),
    }
    series_3['1'].index = pd.date_range(start='2020-01-01', periods=50)
    series_3['2'].index = pd.date_range(start='2020-01-01', periods=30)
    
    exog_2 = {
        '1': exog_wide_range['exog_1'].copy(),
        '2': exog_wide_range['exog_1'].iloc[:30].copy()
    }
    exog_2['1'].index = pd.date_range(start='2020-01-01', periods=50)
    exog_2['2'].index = pd.date_range(start='2020-01-01', periods=30)
    exog_2_pred = {
        '1': exog_pred_wide_range['exog_1'].copy()
    }
    exog_2_pred['1'].index = pd.date_range(start='2020-02-20', periods=50)

    forecaster = ForecasterRecursiveMultiSeries(LinearRegression(), lags=5)
    forecaster.fit(series=series_3, exog=exog_2)

    warn_msg = re.escape(
        "Only series whose last window ends at the same index "
        "can be predicted together. Series that do not reach the "
        "maximum index, '2020-02-19 00:00:00', are excluded "
        "from prediction: {'2'}."
    )
    with pytest.warns(IgnoredArgumentWarning, match = warn_msg):
        predictions = forecaster.predict(steps=5, levels=['1', '2'], last_window=None,
                                         exog = exog_2_pred,)

    expected = pd.DataFrame(
                   data    = np.array([50., 51., 52., 53., 54.]),
                   index   = pd.date_range(start='2020-02-20', periods=5),
                   columns = ['1']
               )

    expected = expected_df_to_long_format(expected)

    pd.testing.assert_frame_equal(predictions, expected)


@pytest.fixture(params=[('1', [50., 51., 52., 53., 54.]), 
                        (['2'], [100., 101., 102., 103., 104.]),
                        (['1', '2'], [[50., 100.],
                                      [51., 101.],
                                      [52., 102.],
                                      [53., 103.],
                                      [54., 104.]])
                        ],
                        ids=lambda d: f'levels: {d[0]}, preds: {d[1]}')
def expected_pandas_dataframe(request):
    """
    This is a pytest fixture. It's a function that can be passed to a
    test so that we have a single block of code that can generate testing
    examples.

    We're using `params` in the call to declare that we want multiple versions
    to be generated. This is similar to the parametrize decorator, but it's difference
    because we can re-use `pd.Series` in multiple tests.
    """
    levels = request.param[0]
    levels_names = [levels] if isinstance(levels, str) else levels
        
    expected_df = pd.DataFrame(
                      data    = request.param[1],
                      columns = levels_names,
                      index   = pd.RangeIndex(start=50, stop=55, step=1)
                  )
    expected_df = expected_df_to_long_format(expected_df)

    return levels, expected_df


def test_predict_output_when_with_fixture(expected_pandas_dataframe):
    """
    Test predict output when using LinearRegression as estimator with pytest fixture.
    This test is equivalent to the next one.
    """
    forecaster = ForecasterRecursiveMultiSeries(LinearRegression(), lags=5)
    forecaster.fit(series=series_2)
    predictions = forecaster.predict(steps=5, levels=expected_pandas_dataframe[0])

    expected = expected_pandas_dataframe[1]

    pd.testing.assert_frame_equal(predictions, expected)


def test_predict_output_when_estimator_is_LinearRegression():
    """
    Test predict output when using LinearRegression as estimator.
    """
    forecaster = ForecasterRecursiveMultiSeries(LinearRegression(), lags=5)
    forecaster.fit(series=series_2)
    predictions_1 = forecaster.predict(steps=5, levels='1')
    expected_1 = pd.DataFrame(
                     data    = np.array([50., 51., 52., 53., 54.]),
                     index   = pd.RangeIndex(start=50, stop=55, step=1),
                     columns = ['1']
                 )

    predictions_2 = forecaster.predict(steps=5, levels=['2'])
    expected_2 = pd.DataFrame(
                     data    = np.array([100., 101., 102., 103., 104.]),
                     index   = pd.RangeIndex(start=50, stop=55, step=1),
                     columns = ['2']
                 )

    predictions_3 = forecaster.predict(steps=5, levels=None)
    expected_3 = pd.DataFrame(
                     data    = np.array([[50., 100.],
                                         [51., 101.],
                                         [52., 102.],
                                         [53., 103.],
                                         [54., 104.]]),
                     index   = pd.RangeIndex(start=50, stop=55, step=1),
                     columns = ['1', '2']
                 )
    
    expected = [expected_1, expected_2, expected_3]
    for i, df in enumerate(expected):
        expected[i] = expected_df_to_long_format(df)

    pd.testing.assert_frame_equal(predictions_1, expected[0])
    pd.testing.assert_frame_equal(predictions_2, expected[1])
    pd.testing.assert_frame_equal(predictions_3, expected[2])


def test_predict_output_when_with_last_window():
    """
    Test predict output when using LinearRegression as estimator and last_window.
    """
    last_window = pd.DataFrame(
                      {'1': [45, 46, 47, 48, 49], 
                       '2': [95, 96, 97, 98, 99], 
                       '3': [1, 2, 3, 4, 5]}, 
                      index = pd.RangeIndex(start=45, stop=50, step=1)
                  )

    forecaster = ForecasterRecursiveMultiSeries(LinearRegression(), lags=5)
    forecaster.fit(series=series_2)
    predictions_1 = forecaster.predict(steps=5, levels='1', last_window=last_window)
    expected_1 = pd.DataFrame(
                     data    = np.array([50., 51., 52., 53., 54.]),
                     index   = pd.RangeIndex(start=50, stop=55, step=1),
                     columns = ['1']
                 )

    predictions_2 = forecaster.predict(steps=5, levels=['2'], last_window=last_window)
    expected_2 = pd.DataFrame(
                     data    = np.array([100., 101., 102., 103., 104.]),
                     index   = pd.RangeIndex(start=50, stop=55, step=1),
                     columns = ['2']
                 )

    predictions_3 = forecaster.predict(steps=5, levels=['1', '2'], last_window=last_window)
    expected_3 = pd.DataFrame(
                     data    = np.array([[50., 100.],
                                         [51., 101.],
                                         [52., 102.],
                                         [53., 103.],
                                         [54., 104.]]),
                     index   = pd.RangeIndex(start=50, stop=55, step=1),
                     columns = ['1', '2']
                 )

    predictions_4 = forecaster.predict(steps=5, levels=None, last_window=last_window[['1', '2']])
    expected_4 = pd.DataFrame(
                     data    = np.array([[50., 100.],
                                         [51., 101.],
                                         [52., 102.],
                                         [53., 103.],
                                         [54., 104.]]),
                     index   = pd.RangeIndex(start=50, stop=55, step=1),
                     columns = ['1', '2']
                 )
    
    expected = [expected_1, expected_2, expected_3, expected_4]
    for i, df in enumerate(expected):
        expected[i] = expected_df_to_long_format(df)

    pd.testing.assert_frame_equal(predictions_1, expected[0])
    pd.testing.assert_frame_equal(predictions_2, expected[1])
    pd.testing.assert_frame_equal(predictions_3, expected[2])
    pd.testing.assert_frame_equal(predictions_4, expected[3])


def test_predict_output_when_with_transform_series():
    """
    Test predict output when using LinearRegression as estimator and StandardScaler.
    """
    forecaster = ForecasterRecursiveMultiSeries(
                     estimator          = LinearRegression(),
                     lags               = 5,
                     transformer_series = StandardScaler()
                 )
    forecaster.fit(series=series_dict_range)
    predictions = forecaster.predict(steps=5, levels='l1')

    expected = pd.DataFrame(
                   data    = np.array([0.52791431, 0.44509712, 0.42176045, 0.48087237, 0.48268008]),
                   index   = pd.RangeIndex(start=50, stop=55, step=1),
                   columns = ['l1']
               )

    expected = expected_df_to_long_format(expected)
    
    pd.testing.assert_frame_equal(predictions, expected)


def test_predict_output_when_with_transform_series_as_dict():
    """
    Test predict output when using LinearRegression as estimator and transformer_series
    is a dict with 2 different transformers.
    """
    transformer_series = {
        'l1': StandardScaler(), 'l2': MinMaxScaler(), '_unknown_level': StandardScaler()
    }
    forecaster = ForecasterRecursiveMultiSeries(
                     estimator          = LinearRegression(),
                     lags               = 5,
                     transformer_series = transformer_series
                 )
    forecaster.fit(series=series_dict_range)
    predictions = forecaster.predict(steps=5, levels=['l1'], suppress_warnings=True)

    expected = pd.DataFrame(
                   data    = np.array([0.59619193, 0.46282914, 0.41738496, 0.48522676, 0.47525733]),
                   index   = pd.RangeIndex(start=50, stop=55, step=1),
                   columns = ['l1']
               )

    expected = expected_df_to_long_format(expected)
    
    pd.testing.assert_frame_equal(predictions, expected)


@pytest.mark.parametrize("transformer_series", 
                         [StandardScaler(),
                          {'l1': StandardScaler(), 'l2': StandardScaler(), '_unknown_level': StandardScaler()}], 
                         ids = lambda tr: f'transformer_series type: {type(tr)}')
def test_predict_output_when_with_transform_series_and_transform_exog(transformer_series):
    """
    Test predict output when using LinearRegression as estimator, StandardScaler
    as transformer_series and transformer_exog as transformer_exog.
    """
    transformer_exog = ColumnTransformer(
                           [('scale', StandardScaler(), ['exog_1']),
                            ('onehot', OneHotEncoder(), ['exog_2'])],
                           remainder = 'passthrough',
                           verbose_feature_names_out = False
                       )
    forecaster = ForecasterRecursiveMultiSeries(
                     estimator          = LinearRegression(),
                     lags               = 5,
                     transformer_series = transformer_series,
                     transformer_exog   = transformer_exog,
                 )
    forecaster.fit(series=series_dict_dt, exog=exog_long_dt)
    predictions = forecaster.predict(steps=5, levels='l1', exog=exog_pred_long_dt)

    expected = pd.DataFrame(
                   data    = np.array([0.53267333, 0.44478046, 0.52579563, 0.57391142, 0.54633594]),
                   index   = pd.date_range(start='2000-02-20', periods=5, freq='D'),
                   columns = ['l1']
               )

    expected = expected_df_to_long_format(expected)
    
    pd.testing.assert_frame_equal(predictions, expected)


@pytest.mark.parametrize("transformer_series", 
                         [StandardScaler(),
                          {'1': StandardScaler(), '2': StandardScaler(), '_unknown_level': StandardScaler()}], 
                         ids = lambda tr: f'transformer_series type: {type(tr)}')
def test_predict_output_when_with_transform_series_and_transform_exog_different_length_series(transformer_series):
    """
    Test predict output when using LinearRegression as estimator, StandardScaler
    as transformer_series and transformer_exog as transformer_exog with series 
    of different lengths.
    """
    series_dict_range_nan = {
        '1': series_dict_range['l1'].copy(),
        '2': series_dict_range['l2'].copy()
    }
    series_dict_range_nan['2'].iloc[:10] = np.nan

    transformer_exog = ColumnTransformer(
                           [('scale', StandardScaler(), ['exog_1']),
                            ('onehot', OneHotEncoder(), ['exog_2'])],
                           remainder = 'passthrough',
                           verbose_feature_names_out = False
                       )
    forecaster = ForecasterRecursiveMultiSeries(
                     estimator          = LinearRegression(),
                     lags               = 5,
                     transformer_series = transformer_series,
                     transformer_exog   = transformer_exog,
                 )
    forecaster.fit(series=series_dict_range_nan, exog=exog_wide_range)
    predictions = forecaster.predict(steps=5, exog=exog_pred_wide_range)

    expected = pd.DataFrame(
                   data    = np.array([[0.52343425, 0.49560898],
                                       [0.42697026, 0.53289813],
                                       [0.48906322, 0.61304224],
                                       [0.55395581, 0.61509108],
                                       [0.52349107, 0.54297301]]),
                   index   = pd.RangeIndex(start=50, stop=55, step=1),
                   columns = ['1', '2']
               )

    expected = expected_df_to_long_format(expected)
    
    pd.testing.assert_frame_equal(predictions, expected)


def test_predict_output_when_categorical_features_native_implementation_HistGradientBoostingRegressor():
    """
    Test predict output when using HistGradientBoostingRegressor and categorical variables.
    """
    df_exog = pd.DataFrame(
        {'exog_1': exog_wide_range['exog_1'],
         'exog_2': ['a', 'b', 'c', 'd', 'e'] * 10,
         'exog_3': pd.Categorical(['F', 'G', 'H', 'I', 'J'] * 10)}
    )
    
    exog_predict = df_exog.copy()
    exog_predict.index = pd.RangeIndex(start=50, stop=100)

    categorical_features = df_exog.select_dtypes(exclude=[np.number]).columns.tolist()
    transformer_exog = make_column_transformer(
                           (
                               OrdinalEncoder(
                                   dtype=int,
                                   handle_unknown="use_encoded_value",
                                   unknown_value=-1,
                                   encoded_missing_value=-1
                               ),
                               categorical_features
                           ),
                           remainder="passthrough",
                           verbose_feature_names_out=False,
                       ).set_output(transform="pandas")
    
    forecaster = ForecasterRecursiveMultiSeries(
                     estimator          = HistGradientBoostingRegressor(
                                              categorical_features = categorical_features,
                                              random_state         = 123
                                          ),
                     lags               = 5,
                     transformer_series = None,
                     transformer_exog   = transformer_exog
                 )
    forecaster.fit(series=series_dict_range, exog=df_exog)
    predictions = forecaster.predict(steps=10, exog=exog_predict)

    expected = pd.DataFrame(
                   data = np.array([[0.55674244, 0.54564144],
                                    [0.39860559, 0.53774558],
                                    [0.53719969, 0.63668042],
                                    [0.7759386 , 0.66871809],
                                    [0.56654221, 0.80858451],
                                    [0.43373393, 0.39998025],
                                    [0.49776999, 0.64896715],
                                    [0.47797067, 0.43578494],
                                    [0.49149785, 0.37087107],
                                    [0.68189581, 0.52863857]]),
                   index   = pd.RangeIndex(start=50, stop=60, step=1),
                   columns = ['l1', 'l2']
               )

    expected = expected_df_to_long_format(expected)
    
    pd.testing.assert_frame_equal(predictions, expected)


def test_predict_output_when_categorical_features_native_implementation_LGBMRegressor():
    """
    Test predict output when using LGBMRegressor and categorical variables.
    """
    df_exog = pd.DataFrame(
        {'exog_1': exog_wide_range['exog_1'],
         'exog_2': ['a', 'b', 'c', 'd', 'e'] * 10,
         'exog_3': pd.Categorical(['F', 'G', 'H', 'I', 'J'] * 10)}
    )

    exog_predict = df_exog.copy()
    exog_predict.index = pd.RangeIndex(start=50, stop=100)

    categorical_features = df_exog.select_dtypes(exclude=[np.number]).columns.tolist()
    transformer_exog = make_column_transformer(
                           (
                               OrdinalEncoder(
                                   dtype=int,
                                   handle_unknown="use_encoded_value",
                                   unknown_value=-1,
                                   encoded_missing_value=-1
                               ),
                               categorical_features
                           ),
                           remainder="passthrough",
                           verbose_feature_names_out=False,
                       ).set_output(transform="pandas")

    forecaster = ForecasterRecursiveMultiSeries(
                     estimator          = LGBMRegressor(random_state=123),
                     lags               = 5,
                     transformer_series = None,
                     transformer_exog   = transformer_exog,
                     fit_kwargs         = {'categorical_feature': categorical_features}
                 )
    forecaster.fit(series=series_dict_range, exog=df_exog)
    predictions = forecaster.predict(steps=10, exog=exog_predict)

    expected = pd.DataFrame(
                   data = np.array([[0.6211422 , 0.73248662],
                                    [0.42283865, 0.45391191],
                                    [0.49351412, 0.661996  ],
                                    [0.71846231, 0.58080355],
                                    [0.55089719, 0.59780378],
                                    [0.39224631, 0.40316854],
                                    [0.46827996, 0.51400857],
                                    [0.67707084, 0.42834292],
                                    [0.45119292, 0.4503941 ],
                                    [0.61998977, 0.65552498]]),
                   index   = pd.RangeIndex(start=50, stop=60, step=1),
                   columns = ['l1', 'l2']
               )

    expected = expected_df_to_long_format(expected)

    pd.testing.assert_frame_equal(predictions, expected)


def test_predict_output_when_categorical_features_native_implementation_LGBMRegressor_auto():
    """
    Test predict output when using LGBMRegressor and categorical variables with 
    categorical_features='auto'.
    """
    df_exog = pd.DataFrame(
        {'exog_1': exog_wide_range['exog_1'],
         'exog_2': ['a', 'b', 'c', 'd', 'e'] * 10,
         'exog_3': pd.Categorical(['F', 'G', 'H', 'I', 'J'] * 10)}
    )
    
    exog_predict = df_exog.copy()
    exog_predict.index = pd.RangeIndex(start=50, stop=100)

    pipeline_categorical = make_pipeline(
                               OrdinalEncoder(
                                   dtype=int,
                                   handle_unknown="use_encoded_value",
                                   unknown_value=-1,
                                   encoded_missing_value=-1
                               ),
                               FunctionTransformer(
                                   func=lambda x: x.astype('category'),
                                   feature_names_out= 'one-to-one'
                               )
                           )
    transformer_exog = make_column_transformer(
                            (
                                pipeline_categorical,
                                make_column_selector(dtype_exclude=np.number)
                            ),
                            remainder="passthrough",
                            verbose_feature_names_out=False,
                       ).set_output(transform="pandas")
    
    forecaster = ForecasterRecursiveMultiSeries(
                     estimator          = LGBMRegressor(random_state=123),
                     lags               = 5,
                     transformer_series = None,
                     transformer_exog   = transformer_exog,
                     fit_kwargs         = {'categorical_feature': 'auto'}
                 )
    forecaster.fit(series=series_dict_range, exog=df_exog)
    predictions = forecaster.predict(steps=10, exog=exog_predict)

    expected = pd.DataFrame(
                   data = np.array([[0.6211422 , 0.73248662],
                                    [0.42283865, 0.45391191],
                                    [0.49351412, 0.661996  ],
                                    [0.71846231, 0.58080355],
                                    [0.55089719, 0.59780378],
                                    [0.39224631, 0.40316854],
                                    [0.46827996, 0.51400857],
                                    [0.67707084, 0.42834292],
                                    [0.45119292, 0.4503941 ],
                                    [0.61998977, 0.65552498]]),
                   index   = pd.RangeIndex(start=50, stop=60, step=1),
                   columns = ['l1', 'l2']
               )

    expected = expected_df_to_long_format(expected)
    
    pd.testing.assert_frame_equal(predictions, expected)


def test_predict_output_when_window_features():
    """
    Test output of predict when estimator is LGBMRegressor and window features.
    """

    rolling = RollingFeatures(stats=['mean', 'median'], window_sizes=4)
    transformer_exog = ColumnTransformer(
                           [('scale', StandardScaler(), ['exog_1']),
                            ('onehot', OneHotEncoder(), ['exog_2'])],
                           remainder = 'passthrough',
                           verbose_feature_names_out = False
                       )
    
    forecaster = ForecasterRecursiveMultiSeries(
                     estimator          = LGBMRegressor(verbose=-1),
                     lags               = 5,
                     window_features    = rolling,
                     transformer_series = StandardScaler(),
                     transformer_exog   = transformer_exog
                 )
    forecaster.fit(series=series_dict_range, exog=exog_wide_range)
    predictions = forecaster.predict(steps=5, exog=exog_pred_wide_range)

    expected = pd.DataFrame(
                   data = np.array([
                              [0.62360244, 0.72875212],
                              [0.47067417, 0.29193681],
                              [0.56289914, 0.60510683],
                              [0.72004167, 0.74123332],
                              [0.51581126, 0.54273085]]),
                   index   = pd.RangeIndex(start=50, stop=55, step=1),
                   columns = ['l1', 'l2']
               )

    expected = expected_df_to_long_format(expected)
    
    pd.testing.assert_frame_equal(predictions, expected)


def test_predict_output_when_series_and_exog_dict():
    """
    Test output ForecasterRecursiveMultiSeries predict method when series and 
    exog are dictionaries.
    """
    forecaster = ForecasterRecursiveMultiSeries(
                     estimator          = LGBMRegressor(
                         n_estimators=2, random_state=123, verbose=-1, max_depth=2
                     ),
                     lags               = 14,
                     encoding           = 'ordinal',
                     dropna_from_series = False,
                     transformer_series = StandardScaler(),
                     transformer_exog   = StandardScaler()
                 )
    forecaster.fit(
        series=series_dict_nans_train, exog=exog_dict_nans_train, suppress_warnings=True
    )
    predictions = forecaster.predict(
        steps=5, exog=exog_dict_nans_test, suppress_warnings=True
    )
    expected = pd.DataFrame(
        data=np.array(
            [
                [1438.14154717, 2090.79352613, 2166.9832933, 7285.52781428],
                [1438.14154717, 2089.11038884, 2074.55994929, 7488.18398744],
                [1438.14154717, 2089.11038884, 2035.99448247, 7488.18398744],
                [1403.93625654, 2089.11038884, 2035.99448247, 7488.18398744],
                [1403.93625654, 2089.11038884, 2035.99448247, 7488.18398744],
            ]
        ),
        index=pd.date_range(start="2016-08-01", periods=5, freq="D"),
        columns=["id_1000", "id_1001", "id_1003", "id_1004"],
    )

    expected = expected_df_to_long_format(expected)

    pd.testing.assert_frame_equal(predictions, expected)


@pytest.mark.parametrize("differentiation", 
                         [1, {'1': 1, '2': 1, '_unknown_level': 1}], 
                         ids = lambda diff: f'differentiation: {diff}')
def test_predict_output_when_with_exog_and_differentiation_is_1(differentiation):
    """
    Test predict output when using LinearRegression as estimator and differentiation=1.
    """    
    series = series_wide_dt.copy()
    series.index = pd.RangeIndex(start=0, stop=len(series), step=1)

    end_train = 30
    # Data differentiated
    differentiator = TimeSeriesDifferentiator(order=1)
    series_1_diff = differentiator.fit_transform(series['1'].to_numpy())
    series_2_diff = differentiator.fit_transform(series['2'].to_numpy())
    data_diff = pd.DataFrame(
        {'1': series_1_diff, '2': series_2_diff}
    ).dropna()
    data_diff.index = series.index[1:]

    # Simulated exogenous variable
    rng = np.random.default_rng(9876)
    exog = pd.Series(
        rng.normal(loc=0, scale=1, size=len(series)), index=series.index, name='exog'
    )
    exog_diff = exog.iloc[1:]

    steps = len(data_diff.loc[end_train + 1:])

    series_dict_train_diff = data_diff.loc[:end_train].to_dict(orient='series')
    forecaster_1 = ForecasterRecursiveMultiSeries(
        estimator=LinearRegression(), lags=15, transformer_series=None
    )
    forecaster_1.fit(series=series_dict_train_diff, exog=exog_diff.loc[:end_train])
    predictions_diff = forecaster_1.predict(steps=steps, exog=exog_diff.loc[end_train + 1:])
    # Revert the differentiation
    predictions_diff = pd.pivot_table(
        predictions_diff, index=predictions_diff.index, columns="level", values="pred"
    )
    last_value_train = series.loc[:end_train].iloc[[-1]]
    predictions_1 = pd.concat([last_value_train, predictions_diff]).cumsum()[1:]
    predictions_1 = expected_df_to_long_format(predictions_1)

    series_train_dict = series.loc[:end_train].to_dict(orient='series')
    forecaster_2 = ForecasterRecursiveMultiSeries(
        estimator=LinearRegression(), lags=15, transformer_series=None, differentiation=differentiation, 
    )
    forecaster_2.fit(series=series_train_dict, exog=exog.loc[:end_train])
    predictions_2 = forecaster_2.predict(steps=steps, exog=exog.loc[end_train + 1:])

    pd.testing.assert_frame_equal(predictions_1, predictions_2)


@pytest.mark.parametrize("differentiation", 
                         [1, {'1': 1, '2': 1, '_unknown_level': 1}], 
                         ids = lambda diff: f'differentiation: {diff}')
def test_predict_output_when_with_exog_differentiation_is_1_and_transformer_series(differentiation):
    """
    Test predict output when using LinearRegression as estimator and differentiation=1,
    and transformer_series is StandardScaler.
    """
    end_train = '2003-01-30 23:59:00'

    # Data scaled and differentiated
    series_datetime = series_wide_dt.copy()
    series_datetime.index = pd.date_range(
        start='2003-01-01', periods=len(series_datetime), freq='D'
    )
    series_dict_datetime = {
        "1": series_datetime['1'].loc[:end_train],
        "2": series_datetime['2'].loc[:end_train]
    }

    scaler_dict = {
        "1": StandardScaler(),
        "2": StandardScaler()
    }

    series_dict_diff = {}
    df_last_value_train = pd.DataFrame()
    for k, v in series_dict_datetime.items():

        scaler_dict[k].fit(v.to_numpy().reshape(-1, 1))
        data_scaled = scaler_dict[k].transform(v.copy().to_numpy().reshape(-1, 1))
        data_scaled = pd.Series(data_scaled.ravel(), index=v.index)
        last_value_train = data_scaled.iloc[[-1]]
        df_last_value_train[k] = last_value_train

        data_scaled_diff = TimeSeriesDifferentiator(order=1).fit_transform(data_scaled.to_numpy())
        data_scaled_diff = pd.Series(data_scaled_diff).dropna()
        data_scaled_diff.index = v.index[1:]
        series_dict_diff[k] = data_scaled_diff
    
    # Simulated exogenous variable
    rng = np.random.default_rng(9876)
    exog = pd.Series(
        rng.normal(loc=0, scale=1, size=len(series_wide_dt)), name='exog'
    )
    exog.index = pd.date_range(start='2003-01-01', periods=len(exog), freq='D')
    exog_dict_datetime = {
        "1": exog.loc[:end_train],
        "2": exog.loc[:end_train]
    }
    exog_diff = {
        '1': exog_dict_datetime['1'].iloc[1:],
        '2': exog_dict_datetime['2'].iloc[1:]
    }
    exog_pred = {
        '1': exog.loc[end_train:],
        '2': exog.loc[end_train:]
    }

    steps = len(series_datetime.loc[end_train:])

    forecaster_1 = ForecasterRecursiveMultiSeries(
        estimator=LinearRegression(), lags=15, transformer_series=None
    )
    forecaster_1.fit(series=series_dict_diff, exog=exog_diff)
    predictions_diff = forecaster_1.predict(steps=steps, exog=exog_pred)
    # Revert the differentiation
    predictions_diff = pd.pivot_table(
        predictions_diff, index=predictions_diff.index, columns="level", values="pred"
    )
    predictions_1 = pd.concat([df_last_value_train, predictions_diff]).cumsum()[1:]
    # Revert the scaling
    for k in scaler_dict.keys():
        predictions_1[k] = scaler_dict[k].inverse_transform(predictions_1[k].to_numpy().reshape(-1, 1))
    predictions_1 = expected_df_to_long_format(predictions_1)

    forecaster_2 = ForecasterRecursiveMultiSeries(
        estimator=LinearRegression(), lags=15, transformer_series=StandardScaler(), differentiation=differentiation
    )
    forecaster_2.fit(series=series_dict_datetime, exog=exog_dict_datetime)
    predictions_2 = forecaster_2.predict(steps=steps, exog=exog_pred)

    pd.testing.assert_frame_equal(predictions_1, predictions_2)


@pytest.mark.parametrize("differentiation", 
                         [2, {'1': 2, '2': 2, '_unknown_level': 2}], 
                         ids = lambda diff: f'differentiation: {diff}')
def test_predict_output_when_with_exog_and_differentiation_is_2(differentiation):
    """
    Test predict output when using LinearRegression as estimator and differentiation=2.
    """
    series_dt = series_wide_dt.copy()
    series_dt.index = pd.date_range(start='2003-01-01', periods=len(series_dt), freq='D')

    # Data differentiated
    df_diff_1 = pd.DataFrame()
    df_diff_2 = pd.DataFrame()
    for col in series_dt.columns:

        differentiator_1 = TimeSeriesDifferentiator(order=1)
        differentiator_2 = TimeSeriesDifferentiator(order=2)

        data_diff_1 = differentiator_1.fit_transform(series_dt[col].to_numpy())
        data_diff_1 = pd.Series(data_diff_1).dropna()
        df_diff_1[col] = data_diff_1

        data_diff_2 = differentiator_2.fit_transform(series_dt[col].to_numpy())
        data_diff_2 = pd.Series(data_diff_2).dropna()
        df_diff_2[col] = data_diff_2

    df_diff_1.index = series_dt.index[1:]
    df_diff_2.index = series_dt.index[2:]

    # Simulated exogenous variable
    rng = np.random.default_rng(9876)
    exog = pd.Series(
        rng.normal(loc=0, scale=1, size=len(series_dt)), index=series_dt.index, name='exog'
    )
    exog_diff_2 = exog.iloc[2:]

    end_train = '2003-01-30 23:59:00'
    steps = len(series_dt.loc[end_train:])

    series_dict_train_diff_2 = df_diff_2.loc[:end_train].to_dict(orient='series')
    forecaster_1 = ForecasterRecursiveMultiSeries(
        estimator=LinearRegression(), lags=15, transformer_series=None
    )
    forecaster_1.fit(series=series_dict_train_diff_2, exog=exog_diff_2.loc[:end_train])
    predictions_diff_2 = forecaster_1.predict(steps=steps, exog=exog_diff_2.loc[end_train:])
    
    # Revert the differentiation
    predictions_diff_2 = pd.pivot_table(
        predictions_diff_2, index=predictions_diff_2.index, columns="level", values="pred"
    )
    last_value_train_diff = df_diff_1.loc[:end_train].iloc[[-1]]
    predictions_diff_1 = pd.concat([last_value_train_diff, predictions_diff_2]).cumsum()[1:]
    last_value_train = series_dt.loc[:end_train].iloc[[-1]]
    predictions_1 = pd.concat([last_value_train, predictions_diff_1]).cumsum()[1:]
    predictions_1 = expected_df_to_long_format(predictions_1)

    series_train_dict = series_dt.loc[:end_train].to_dict(orient='series')
    forecaster_2 = ForecasterRecursiveMultiSeries(
        estimator=LinearRegression(), lags=15, transformer_series=None, differentiation=differentiation
    )
    forecaster_2.fit(series=series_train_dict, exog=exog.loc[:end_train])
    predictions_2 = forecaster_2.predict(steps=steps, exog=exog.loc[end_train:])

    pd.testing.assert_frame_equal(predictions_1, predictions_2)


@pytest.mark.parametrize("levels", 
                         [None, ['id_1000', 'id_1001', 'id_1003', 'id_1004', 'id_1005']], 
                         ids = lambda levels: f'levels: {levels}')
def test_predict_output_when_series_and_exog_dict_encoding_None_unknown_level(levels):
    """
    Test output ForecasterRecursiveMultiSeries predict method when series and 
    exog are dictionaries and encoding is None and unknown level with no exog.
    """
    forecaster = ForecasterRecursiveMultiSeries(
                     estimator          = LGBMRegressor(
                         n_estimators=30, random_state=123, verbose=-1, max_depth=4
                     ),
                     lags               = [1, 7, 14],
                     encoding           = None,
                     dropna_from_series = False,
                     transformer_series = None,
                     transformer_exog   = StandardScaler(),
                     differentiation    = 1
                 )
    forecaster.fit(
        series=series_dict_nans_train, exog=exog_dict_nans_train, suppress_warnings=True
    )

    series_to_predict = ['id_1000', 'id_1001', 'id_1003', 'id_1004', 'id_1005']
    last_window = pd.DataFrame(
        {k: v for k, v in forecaster.last_window_.items() if k in series_to_predict}
    )
    last_window['id_1005'] = last_window['id_1004'] * 0.9
    predictions = forecaster.predict(
        steps=5, levels=levels, last_window=last_window,
        exog=exog_dict_nans_test, suppress_warnings=True
    )

    expected = pd.DataFrame(
        data=np.array([
            [1351.44874045, 3267.13419659, 3843.06135374, 7220.15909652, 5776.91782212],
            [1411.73786344, 3537.21728977, 3823.777024  , 7541.54119992, 6083.61461454],
            [1367.56233886, 3595.37617537, 3861.96220571, 7599.93110962, 6018.65643062],
            [1349.98221742, 3730.36025071, 3900.14738742, 7997.86353537, 6213.61087632],
            [1330.33321825, 3752.7395927 , 3850.39111741, 7923.44398373, 6128.67898851]]),
        index=pd.date_range(start="2016-08-01", periods=5, freq="D"),
        columns=["id_1000", "id_1001", "id_1003", "id_1004", "id_1005"],
    )

    expected = expected_df_to_long_format(expected)

    pd.testing.assert_frame_equal(predictions, expected)


@pytest.mark.parametrize("levels", 
                         [None, ['id_1000', 'id_1001', 'id_1003', 'id_1004', 'id_1005']], 
                         ids = lambda levels: f'levels: {levels}')
def test_predict_output_when_series_and_exog_dict_encoding_None_transformer_series_unknown_level(levels):
    """
    Test output ForecasterRecursiveMultiSeries predict method when series and 
    exog are dictionaries and encoding is None and transformer_series is StandardScaler.
    """
    forecaster = ForecasterRecursiveMultiSeries(
                     estimator          = LGBMRegressor(
                         n_estimators=30, random_state=123, verbose=-1, max_depth=4
                     ),
                     lags               = 14,
                     encoding           = None,
                     dropna_from_series = False,
                     transformer_series = StandardScaler(),
                     transformer_exog   = StandardScaler()
                 )
    forecaster.fit(
        series=series_dict_nans_train, exog=exog_dict_nans_train, suppress_warnings=True
    )

    series_to_predict = ['id_1000', 'id_1001', 'id_1003', 'id_1004', 'id_1005']
    last_window = pd.DataFrame(
        {k: v for k, v in forecaster.last_window_.items() if k in series_to_predict}
    )
    last_window['id_1005'] = last_window['id_1004'] * 0.9
    exog_dict_test_2 = exog_dict_nans_test.copy()
    exog_dict_test_2['id_1005'] = exog_dict_test_2['id_1001']
    predictions = forecaster.predict(
        steps=5, levels=levels, last_window=last_window,
        exog=exog_dict_test_2, suppress_warnings=True
    )
    expected = pd.DataFrame(
        data=np.array([
            [1540.61279683, 3160.95721541, 3121.57888505, 7369.19869868, 7011.48712882],
            [1533.43616081, 3157.30206473, 2976.43108071, 8430.97835738, 8401.89546128],
            [1466.94536267, 3215.9383148 , 2991.20843378, 8565.22794876, 8506.18236246],
            [1427.46483497, 3512.06387649, 2806.20216519, 8343.01317975, 8561.27068389],
            [1411.73980543, 3275.61791117, 1959.1294473 , 8360.01068926, 7633.82955233]]),
        index=pd.date_range(start="2016-08-01", periods=5, freq="D"),
        columns=["id_1000", "id_1001", "id_1003", "id_1004", "id_1005"],
    )

    expected = expected_df_to_long_format(expected)

    pd.testing.assert_frame_equal(predictions, expected)


@pytest.mark.parametrize("levels", 
                         [None, ['id_1000', 'id_1001', 'id_1003', 'id_1004', 'id_1005']], 
                         ids = lambda levels: f'levels: {levels}')
def test_predict_output_when_series_and_exog_dict_unknown_level(levels):
    """
    Test output ForecasterRecursiveMultiSeries predict method when series and 
    exog are dictionaries and unknown level.
    """
    forecaster = ForecasterRecursiveMultiSeries(
                     estimator          = LGBMRegressor(
                         n_estimators=30, random_state=123, verbose=-1, max_depth=4
                     ),
                     lags               = [1, 7, 14],
                     encoding           = 'onehot',
                     dropna_from_series = False,
                     transformer_series = StandardScaler(),
                     transformer_exog   = StandardScaler()
                 )
    forecaster.fit(
        series=series_dict_nans_train, exog=exog_dict_nans_train, suppress_warnings=True
    )

    series_to_predict = ['id_1000', 'id_1001', 'id_1003', 'id_1004', 'id_1005']
    last_window = pd.DataFrame(
        {k: v for k, v in forecaster.last_window_.items() if k in series_to_predict}
    )
    last_window['id_1005'] = last_window['id_1004'] * 0.9
    exog_dict_test_2 = exog_dict_nans_test.copy()
    exog_dict_test_2['id_1005'] = exog_dict_test_2['id_1001']
    predictions = forecaster.predict(
        steps=5, levels=levels, last_window=last_window,
        exog=exog_dict_test_2, suppress_warnings=True
    )
    expected = pd.DataFrame(
        data=np.array([
            [1444.89638131, 2608.52179991, 2594.53128524, 7607.65586239, 5178.02908923],
            [1426.33025646, 2522.90927921, 2240.71166036, 7993.7648681 , 4793.03739895],
            [1419.2856951 , 2451.51448204, 2183.50817345, 8473.84457748, 4721.84643207],
            [1368.78988907, 2477.40763584, 2180.37278483, 8528.07173563, 4559.15070243],
            [1324.15676752, 2566.91100046, 2186.03652396, 8767.76813577, 4966.9971064 ]]),
        index=pd.date_range(start="2016-08-01", periods=5, freq="D"),
        columns=["id_1000", "id_1001", "id_1003", "id_1004", "id_1005"],
    )

    expected = expected_df_to_long_format(expected)

    pd.testing.assert_frame_equal(predictions, expected)



# Unit test fit ForecasterStats
# ==============================================================================
import re
import pytest
import numpy as np
import pandas as pd
from skforecast.stats import Sarimax, Arar
from skforecast.recursive import ForecasterStats
from skforecast.exceptions import IgnoredArgumentWarning

# Fixtures
from .fixtures_forecaster_stats import y
from .fixtures_forecaster_stats import y_datetime


def test_fit_ValueError_when_len_exog_is_not_the_same_as_len_y():
    """
    Raise ValueError if the length of `exog` is different from the length of `y`.
    """
    y = pd.Series(data=np.arange(10))
    exog = pd.Series(data=np.arange(11))
    forecaster = ForecasterStats(estimator=Sarimax(order=(1, 1, 1)))

    err_msg = re.escape(
        f"`exog` must have same number of samples as `y`. "
        f"length `exog`: ({len(exog)}), length `y`: ({len(y)})"
    )
    with pytest.raises(ValueError, match = err_msg):
        forecaster.fit(y=y, exog=exog)


def test_IgnoredArgumentWarning_when_estimator_does_not_support_exog():
    """
    Test IgnoredArgumentWarning is raised when estimator does not support exog.
    """
    y = pd.Series(data=np.arange(10), name='y')
    exog = pd.Series(data=np.arange(10), name='exog')
    forecaster = ForecasterStats(estimator=Arar())

    warn_msg = re.escape(
        f"The estimator {forecaster.estimator_type} does not support exogenous variables, "
        f"they will be ignored during fit."
    )
    with pytest.warns(IgnoredArgumentWarning, match = warn_msg):
        forecaster.fit(y=y, exog=exog)


def test_forecaster_y_exog_features_stored():
    """
    Test forecaster stores y and exog features after fitting.
    """
    
    y = pd.Series(data=np.arange(10), name='y_sarimax')
    exog = pd.Series(data=np.arange(10), name='exog')
    forecaster = ForecasterStats(estimator=Sarimax(order=(1, 1, 1)))
    forecaster.fit(y=y, exog=exog)

    series_name_in_ = 'y_sarimax'
    exog_in_ = True
    exog_type_in_ = type(exog)
    exog_names_in_ = ['exog']
    exog_dtypes_in_ = {'exog': exog.dtype}
    exog_dtypes_out_ = {'exog': exog.dtype}
    X_train_exog_names_out_ = ['exog']
    
    assert forecaster.series_name_in_ == series_name_in_
    assert forecaster.exog_in_ == exog_in_
    assert forecaster.exog_type_in_ == exog_type_in_
    assert forecaster.exog_names_in_ == exog_names_in_
    assert forecaster.exog_dtypes_in_ == exog_dtypes_in_
    assert forecaster.exog_dtypes_out_ == exog_dtypes_out_
    assert forecaster.X_train_exog_names_out_ == X_train_exog_names_out_


def test_forecaster_DatetimeIndex_index_freq_stored():
    """
    Test serie_with_DatetimeIndex.index.freq is stored in forecaster.index_freq_.
    """
    serie_with_DatetimeIndex = pd.Series(
        data  = [1, 2, 3, 4, 5],
        index = pd.date_range(start='2022-01-01', periods=5)
    )
    forecaster = ForecasterStats(estimator=Sarimax(order=(1, 0, 0)))
    forecaster.fit(y=serie_with_DatetimeIndex)
    expected = serie_with_DatetimeIndex.index.freq
    results = forecaster.index_freq_

    assert results == expected


@pytest.mark.parametrize("suppress_warnings", 
                         [True, False], 
                         ids = lambda v: f'suppress_warnings: {v}')
def test_forecaster_index_step_stored_with_suppress_warnings(suppress_warnings):
    """
    Test serie without DatetimeIndex, step is stored in forecaster.index_freq_.
    """
    y = pd.Series(data=np.arange(10))
    forecaster = ForecasterStats(estimator=Sarimax(order=(1, 0, 0)))
    forecaster.fit(y=y, suppress_warnings=suppress_warnings)
    expected = y.index.step
    results = forecaster.index_freq_

    assert results == expected


@pytest.mark.parametrize("store_last_window", 
                         [True, False], 
                         ids=lambda lw: f'store_last_window: {lw}')
def test_fit_last_window_stored(store_last_window):
    """
    Test that values of last window are stored after fitting.
    """
    forecaster = ForecasterStats(estimator=Sarimax(order=(1, 0, 0)))
    forecaster.fit(y=pd.Series(np.arange(50)), 
                   store_last_window=store_last_window)
    expected = pd.Series(np.arange(50))

    if store_last_window:
        pd.testing.assert_series_equal(forecaster.last_window_, expected)
    else:
        assert forecaster.last_window_ is None


@pytest.mark.parametrize("y          , idx", 
                         [(y         , pd.RangeIndex(start=0, stop=50)), 
                          (y_datetime, pd.date_range(start='2000', periods=50, freq='YE'))], 
                         ids = lambda values: f'y, index: {type(values)}')
def test_fit_extended_index_stored(y, idx):
    """
    Test that values of self.estimator.arima_res_.fittedvalues.index are 
    stored after fitting in forecaster.extended_index_.
    """
    forecaster = ForecasterStats(estimator=Sarimax(order=(1, 0, 0)))
    forecaster.fit(y=y)

    pd.testing.assert_index_equal(forecaster.extended_index_, idx)



# Unit test create_predict_X ForecasterRecursiveClassifier
# ==============================================================================
import re
import pytest
import numpy as np
import pandas as pd
from sklearn.exceptions import NotFittedError
from sklearn.compose import ColumnTransformer
from sklearn.preprocessing import StandardScaler
from sklearn.preprocessing import OneHotEncoder
from sklearn.preprocessing import OrdinalEncoder
from sklearn.preprocessing import FunctionTransformer
from sklearn.compose import make_column_transformer, make_column_selector
from sklearn.pipeline import make_pipeline
from sklearn.linear_model import LogisticRegression
from sklearn.ensemble import HistGradientBoostingClassifier
from lightgbm import LGBMClassifier

from ....exceptions import DataTransformationWarning
from skforecast.utils import transform_numpy
from skforecast.preprocessing import RollingFeaturesClassification
from skforecast.recursive import ForecasterRecursiveClassifier

# Fixtures
from .fixtures_forecaster_recursive_classifier import y, y_dt
from .fixtures_forecaster_recursive_classifier import exog, exog_dt, exog_predict, exog_dt_predict


def test_create_predict_X_NotFittedError_when_fitted_is_False():
    """
    Test NotFittedError is raised when fitted is False.
    """
    forecaster = ForecasterRecursiveClassifier(
                     estimator = LogisticRegression(),
                     lags      = 5
                 )

    err_msg = re.escape(
        "This Forecaster instance is not fitted yet. Call `fit` with "
        "appropriate arguments before using predict."
    )
    with pytest.raises(NotFittedError, match = err_msg):
        forecaster.create_predict_X(steps=5)


def test_create_predict_X_output_when_estimator_is_LogisticRegression():
    """
    Test create_predict_X output when using LogisticRegression as estimator.
    """
    y_dummy = pd.Series(
        np.array(['a', 'b', 'c', 'a', 'b', 'c', 'a', 'b', 'c', 'a', 'b', 'c', 'a', 'b', 'c']), 
        index=pd.date_range("2020-01-01", periods=15),
        name='y'
    )

    forecaster = ForecasterRecursiveClassifier(LogisticRegression(), lags=3)
    forecaster.fit(y=y_dummy)
    predictions = forecaster.create_predict_X(steps=5)

    expected = pd.DataFrame(
                   data = np.array([[2., 1., 0.],
                                    [0., 2., 1.],
                                    [1., 0., 2.],
                                    [2., 1., 0.],
                                    [0., 2., 1.]]),
                   index = pd.date_range("2020-01-16", periods=5),
                   columns = ['lag_1', 'lag_2', 'lag_3']
               )
    
    pd.testing.assert_frame_equal(predictions, expected)


def test_create_predict_X_output_when_with_exog():
    """
    Test create_predict_X output when using LogisticRegression as estimator.
    """
    forecaster = ForecasterRecursiveClassifier(LogisticRegression(), lags=3)
    forecaster.fit(y=y, exog=exog)
    
    predictions = forecaster.create_predict_X(steps=5, exog=exog_predict)

    expected = pd.DataFrame(
        data=np.array(
            [
                [1.0, 2.0, 0.0, 0.12062867],
                [1.0, 1.0, 2.0, 0.8263408],
                [1.0, 1.0, 1.0, 0.60306013],
                [1.0, 1.0, 1.0, 0.54506801],
                [1.0, 1.0, 1.0, 0.34276383],
            ]
        ),
        index=pd.RangeIndex(start=50, stop=55, step=1),
        columns=["lag_1", "lag_2", "lag_3", "exog"],
    )
    
    pd.testing.assert_frame_equal(predictions, expected)


def test_create_predict_X_output_with_transform_exog():
    """
    Test create_predict_X output when using LogisticRegression as estimator and 
    transformer_exog.
    """

    df_exog = pd.DataFrame({
               'col_1': exog.to_numpy(),
               'col_2': ['a', 'a', 'b', 'a', 'b'] * 10}
           )
    df_exog_predict = df_exog.iloc[:10, :].copy()
    df_exog_predict.index = pd.RangeIndex(start=50, stop=60)

    transformer_exog = ColumnTransformer(
                            [('scale', StandardScaler(), ['col_1']),
                             ('onehot', OneHotEncoder(), ['col_2'])],
                            remainder = 'passthrough',
                            verbose_feature_names_out = False
                       )
    forecaster = ForecasterRecursiveClassifier(
                     estimator        = LogisticRegression(),
                     lags             = 5,
                     transformer_exog = transformer_exog,
                 )
    forecaster.fit(y=y, exog=df_exog)

    warn_msg = re.escape(
        "The output matrix is in the transformed scale due to the "
        "inclusion of transformations (`transformer_exog`) in the Forecaster. "
        "As a result, any predictions generated using this matrix will also "
        "be in the transformed scale. Please refer to the documentation "
        "for more details: "
        "https://skforecast.org/latest/user_guides/training-and-prediction-matrices.html"
    )
    with pytest.warns(DataTransformationWarning, match = warn_msg):
        predictions = forecaster.create_predict_X(steps=5, exog=df_exog_predict)

    expected = pd.DataFrame(
        data=np.array(
            [
                [1.0, 2.0, 0.0, 1.0, 1.0, -1.47636391, 1.0, 0.0],
                [2.0, 1.0, 2.0, 0.0, 1.0, 1.26277054, 1.0, 0.0],
                [0.0, 2.0, 1.0, 2.0, 0.0, 0.3961342, 0.0, 1.0],
                [1.0, 0.0, 2.0, 1.0, 2.0, 0.17104495, 1.0, 0.0],
                [2.0, 1.0, 0.0, 2.0, 1.0, -0.61417373, 0.0, 1.0],
            ]
        ),
        index=pd.RangeIndex(start=50, stop=55, step=1),
        columns=[
            "lag_1",
            "lag_2",
            "lag_3",
            "lag_4",
            "lag_5",
            "col_1",
            "col_2_a",
            "col_2_b",
        ],
    )

    pd.testing.assert_frame_equal(predictions, expected)


def test_create_predict_X_output_when_categorical_features_native_implementation_HistGradientBoostingClassifier():
    """
    Test create_predict_X output when using HistGradientBoostingClassifier and categorical variables.
    """
    df_exog = pd.DataFrame({'exog_1': exog.to_numpy(),
                            'exog_2': ['a', 'b', 'c', 'd', 'e'] * 10,
                            'exog_3': pd.Categorical(['F', 'G', 'H', 'I', 'J'] * 10)})

    df_exog_predict = df_exog.iloc[:10, :].copy()
    df_exog_predict.index = pd.RangeIndex(start=50, stop=60)

    categorical_features = df_exog.select_dtypes(exclude=[np.number]).columns.tolist()
    transformer_exog = make_column_transformer(
                           (
                               OrdinalEncoder(
                                   dtype=int,
                                   handle_unknown="use_encoded_value",
                                   unknown_value=-1,
                                   encoded_missing_value=-1
                               ),
                               categorical_features
                           ),
                           remainder="passthrough",
                           verbose_feature_names_out=False,
                       ).set_output(transform="pandas")

    forecaster = ForecasterRecursiveClassifier(
                     estimator        = HistGradientBoostingClassifier(
                                            categorical_features = categorical_features,
                                            random_state         = 123
                                        ),
                     lags             = 5,
                     transformer_exog = transformer_exog
                 )
    forecaster.fit(y=y, exog=df_exog)

    warn_msg = re.escape(
        "The output matrix is in the transformed scale due to the "
        "inclusion of transformations (`transformer_exog`) in the Forecaster. "
        "As a result, any predictions generated using this matrix will also "
        "be in the transformed scale. Please refer to the documentation "
        "for more details: "
        "https://skforecast.org/latest/user_guides/training-and-prediction-matrices.html"
    )
    with pytest.warns(DataTransformationWarning, match = warn_msg):
        predictions = forecaster.create_predict_X(steps=10, exog=df_exog_predict)

    expected = pd.DataFrame(
        data=np.array(
            [
                [1.0, 2.0, 0.0, 1.0, 1.0, 0.0, 0.0, 0.12062867],
                [1.0, 1.0, 2.0, 0.0, 1.0, 1.0, 1.0, 0.8263408],
                [1.0, 1.0, 1.0, 2.0, 0.0, 2.0, 2.0, 0.60306013],
                [1.0, 1.0, 1.0, 1.0, 2.0, 3.0, 3.0, 0.54506801],
                [1.0, 1.0, 1.0, 1.0, 1.0, 4.0, 4.0, 0.34276383],
                [1.0, 1.0, 1.0, 1.0, 1.0, 0.0, 0.0, 0.30412079],
                [1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 0.41702221],
                [1.0, 1.0, 1.0, 1.0, 1.0, 2.0, 2.0, 0.68130077],
                [1.0, 1.0, 1.0, 1.0, 1.0, 3.0, 3.0, 0.87545684],
                [1.0, 1.0, 1.0, 1.0, 1.0, 4.0, 4.0, 0.51042234],
            ]
        ),
        index=pd.RangeIndex(start=50, stop=60, step=1),
        columns=[
            "lag_1",
            "lag_2",
            "lag_3",
            "lag_4",
            "lag_5",
            "exog_2",
            "exog_3",
            "exog_1",
        ],
    )
    for col in ['lag_1', 'lag_2', 'lag_3', 'lag_4', 'lag_5']:
        expected[col] = pd.Categorical(
                            values     = expected[col],
                            categories = forecaster.class_codes_,
                            ordered    = False
                        )

    pd.testing.assert_frame_equal(predictions, expected)


def test_create_predict_X_output_when_categorical_features_native_implementation_LGBMClassifier():
    """
    Test create_predict_X output when using LGBMClassifier and categorical variables.
    """
    df_exog = pd.DataFrame({'exog_1': exog.to_numpy(),
                            'exog_2': ['a', 'b', 'c', 'd', 'e'] * 10,
                            'exog_3': pd.Categorical(['F', 'G', 'H', 'I', 'J'] * 10)})
    
    df_exog_predict = df_exog.iloc[:10, :].copy()
    df_exog_predict.index = pd.RangeIndex(start=50, stop=60)

    categorical_features = df_exog.select_dtypes(exclude=[np.number]).columns.tolist()
    transformer_exog = make_column_transformer(
                           (
                               OrdinalEncoder(
                                   dtype=int,
                                   handle_unknown="use_encoded_value",
                                   unknown_value=-1,
                                   encoded_missing_value=-1
                               ),
                               categorical_features
                           ),
                           remainder="passthrough",
                           verbose_feature_names_out=False,
                       ).set_output(transform="pandas")
    
    forecaster = ForecasterRecursiveClassifier(
                     estimator        = LGBMClassifier(verbose=-1, random_state=123),
                     lags             = 5,
                     transformer_exog = transformer_exog,
                     fit_kwargs       = {'categorical_feature': categorical_features}
                 )
    forecaster.fit(y=y, exog=df_exog)

    warn_msg = re.escape(
        "The output matrix is in the transformed scale due to the "
        "inclusion of transformations (`transformer_exog`) in the Forecaster. "
        "As a result, any predictions generated using this matrix will also "
        "be in the transformed scale. Please refer to the documentation "
        "for more details: "
        "https://skforecast.org/latest/user_guides/training-and-prediction-matrices.html"
    )
    with pytest.warns(DataTransformationWarning, match = warn_msg):
        predictions = forecaster.create_predict_X(steps=10, exog=df_exog_predict)

    expected = pd.DataFrame(
        data=np.array(
            [
                [1.0, 2.0, 0.0, 1.0, 1.0, 0.0, 0.0, 0.12062867],
                [1.0, 1.0, 2.0, 0.0, 1.0, 1.0, 1.0, 0.8263408],
                [1.0, 1.0, 1.0, 2.0, 0.0, 2.0, 2.0, 0.60306013],
                [1.0, 1.0, 1.0, 1.0, 2.0, 3.0, 3.0, 0.54506801],
                [1.0, 1.0, 1.0, 1.0, 1.0, 4.0, 4.0, 0.34276383],
                [1.0, 1.0, 1.0, 1.0, 1.0, 0.0, 0.0, 0.30412079],
                [1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 0.41702221],
                [1.0, 1.0, 1.0, 1.0, 1.0, 2.0, 2.0, 0.68130077],
                [1.0, 1.0, 1.0, 1.0, 1.0, 3.0, 3.0, 0.87545684],
                [1.0, 1.0, 1.0, 1.0, 1.0, 4.0, 4.0, 0.51042234],
            ]
        ),
        index=pd.RangeIndex(start=50, stop=60, step=1),
        columns=[
            "lag_1",
            "lag_2",
            "lag_3",
            "lag_4",
            "lag_5",
            "exog_2",
            "exog_3",
            "exog_1",
        ],
    )
    for col in ['lag_1', 'lag_2', 'lag_3', 'lag_4', 'lag_5']:
        expected[col] = pd.Categorical(
                            values     = expected[col],
                            categories = forecaster.class_codes_,
                            ordered    = False
                        )

    pd.testing.assert_frame_equal(predictions, expected)


def test_create_predict_X_output_when_categorical_features_native_implementation_LGBMClassifier_auto():
    """
    Test create_predict_X output when using LGBMClassifier and categorical variables with 
    categorical_features='auto'.
    """
    df_exog = pd.DataFrame({'exog_1': exog.to_numpy(),
                            'exog_2': ['a', 'b', 'c', 'd', 'e'] * 10,
                            'exog_3': pd.Categorical(['F', 'G', 'H', 'I', 'J'] * 10)})

    df_exog_predict = df_exog.iloc[:10, :].copy()
    df_exog_predict.index = pd.RangeIndex(start=50, stop=60)

    pipeline_categorical = make_pipeline(
                               OrdinalEncoder(
                                   dtype=int,
                                   handle_unknown="use_encoded_value",
                                   unknown_value=-1,
                                   encoded_missing_value=-1
                               ),
                               FunctionTransformer(
                                   func=lambda x: x.astype('category'),
                                   feature_names_out= 'one-to-one'
                               )
                           )
    transformer_exog = make_column_transformer(
                            (
                                pipeline_categorical,
                                make_column_selector(dtype_exclude=np.number)
                            ),
                            remainder="passthrough",
                            verbose_feature_names_out=False,
                       ).set_output(transform="pandas")

    forecaster = ForecasterRecursiveClassifier(
                     estimator        = LGBMClassifier(verbose=-1, random_state=123),
                     lags             = 5,
                     transformer_exog = transformer_exog,
                     fit_kwargs       = {'categorical_feature': 'auto'}
                 )
    forecaster.fit(y=y, exog=df_exog)

    warn_msg = re.escape(
        "The output matrix is in the transformed scale due to the "
        "inclusion of transformations (`transformer_exog`) in the Forecaster. "
        "As a result, any predictions generated using this matrix will also "
        "be in the transformed scale. Please refer to the documentation "
        "for more details: "
        "https://skforecast.org/latest/user_guides/training-and-prediction-matrices.html"
    )
    with pytest.warns(DataTransformationWarning, match = warn_msg):
        predictions = forecaster.create_predict_X(steps=10, exog=df_exog_predict)

    expected = pd.DataFrame(
        data=np.array(
            [
                [1.0, 2.0, 0.0, 1.0, 1.0, 0.0, 0.0, 0.12062867],
                [1.0, 1.0, 2.0, 0.0, 1.0, 1.0, 1.0, 0.8263408],
                [0.0, 1.0, 1.0, 2.0, 0.0, 2.0, 2.0, 0.60306013],
                [1.0, 0.0, 1.0, 1.0, 2.0, 3.0, 3.0, 0.54506801],
                [1.0, 1.0, 0.0, 1.0, 1.0, 4.0, 4.0, 0.34276383],
                [0.0, 1.0, 1.0, 0.0, 1.0, 0.0, 0.0, 0.30412079],
                [0.0, 0.0, 1.0, 1.0, 0.0, 1.0, 1.0, 0.41702221],
                [1.0, 0.0, 0.0, 1.0, 1.0, 2.0, 2.0, 0.68130077],
                [1.0, 1.0, 0.0, 0.0, 1.0, 3.0, 3.0, 0.87545684],
                [0.0, 1.0, 1.0, 0.0, 0.0, 4.0, 4.0, 0.51042234],
            ]
        ),
        index=pd.RangeIndex(start=50, stop=60, step=1),
        columns=[
            "lag_1",
            "lag_2",
            "lag_3",
            "lag_4",
            "lag_5",
            "exog_2",
            "exog_3",
            "exog_1",
        ],
    )
    for col in ['lag_1', 'lag_2', 'lag_3', 'lag_4', 'lag_5']:
        expected[col] = pd.Categorical(
                            values     = expected[col],
                            categories = forecaster.class_codes_,
                            ordered    = False
                        )
    expected['exog_2'] = pd.Categorical(
                             values     = expected['exog_2'],
                             categories = [0, 1, 2, 3, 4],
                             ordered    = False
                         )
    expected['exog_3'] = pd.Categorical(
                             values     = expected['exog_3'],
                             categories = [0, 1, 2, 3, 4],
                             ordered    = False
                         )

    pd.testing.assert_frame_equal(predictions, expected)


@pytest.mark.parametrize("steps", 
                         [10, '2020-02-29', pd.to_datetime('2020-02-29')], 
                         ids=lambda steps: f'steps: {steps}')
def test_create_predict_X_output_when_window_features(steps):
    """
    Test output of create_predict_X when estimator is LGBMClassifier and window features.
    """

    rolling = RollingFeaturesClassification(stats=['proportion', 'entropy'], window_sizes=[3, 5])
    forecaster = ForecasterRecursiveClassifier(
        LGBMClassifier(verbose=-1, random_state=123), lags=3, window_features=rolling
    )
    forecaster.fit(y=y_dt, exog=exog_dt)
    predictions = forecaster.create_predict_X(steps=steps, exog=exog_dt_predict)

    expected = pd.DataFrame(
        data=np.array(
            [
                [
                    1.0,
                    2.0,
                    0.0,
                    0.33333333,
                    0.33333333,
                    0.33333333,
                    1.37095059,
                    0.12062867,
                ],
                [2.0, 1.0, 2.0, 0.0, 0.33333333, 0.66666667, 1.52192809, 0.8263408],
                [1.0, 2.0, 1.0, 0.0, 0.66666667, 0.33333333, 1.52192809, 0.60306013],
                [1.0, 1.0, 2.0, 0.0, 0.66666667, 0.33333333, 0.97095059, 0.54506801],
                [1.0, 1.0, 1.0, 0.0, 1.0, 0.0, 0.72192809, 0.34276383],
                [0.0, 1.0, 1.0, 0.33333333, 0.66666667, 0.0, 1.37095059, 0.30412079],
                [
                    2.0,
                    0.0,
                    1.0,
                    0.33333333,
                    0.33333333,
                    0.33333333,
                    1.37095059,
                    0.41702221,
                ],
                [
                    1.0,
                    2.0,
                    0.0,
                    0.33333333,
                    0.33333333,
                    0.33333333,
                    1.37095059,
                    0.68130077,
                ],
                [1.0, 1.0, 2.0, 0.0, 0.66666667, 0.33333333, 1.37095059, 0.87545684],
                [0.0, 1.0, 1.0, 0.33333333, 0.66666667, 0.0, 1.52192809, 0.51042234],
            ]
        ),
        index=pd.date_range("2020-02-20", periods=10),
        columns=[
            "lag_1",
            "lag_2",
            "lag_3",
            "roll_proportion_3_class_0",
            "roll_proportion_3_class_1",
            "roll_proportion_3_class_2",
            "roll_entropy_5",
            "exog",
        ],
    )
    for col in ['lag_1', 'lag_2', 'lag_3']:
        expected[col] = pd.Categorical(
                            values     = expected[col],
                            categories = forecaster.class_codes_,
                            ordered    = False
                        )

    pd.testing.assert_frame_equal(predictions, expected)




# Unit test _create_lags ForecasterRecursiveClassifier
# ==============================================================================
import numpy as np
import pandas as pd
from lightgbm import LGBMClassifier
from sklearn.linear_model import LogisticRegression
from skforecast.preprocessing import RollingFeaturesClassification
from skforecast.recursive import ForecasterRecursiveClassifier

rolling = RollingFeaturesClassification(
    stats=['proportion', 'mode', 'entropy'], window_sizes=[5, 5, 6]
)


def test_create_lags_output():
    """
    Test matrix of lags is created properly when lags=3 and y=np.arange(10).
    """
    forecaster = ForecasterRecursiveClassifier(LogisticRegression(), lags=3)
    results = forecaster._create_lags(y=np.arange(10))
    expected = (
        np.array([[2., 1., 0.],
                  [3., 2., 1.],
                  [4., 3., 2.],
                  [5., 4., 3.],
                  [6., 5., 4.],
                  [7., 6., 5.],
                  [8., 7., 6.]]),
        np.array([3., 4., 5., 6., 7., 8., 9.])
    )

    np.testing.assert_array_almost_equal(results[0], expected[0])
    np.testing.assert_array_almost_equal(results[1], expected[1])


def test_create_lags_output_interspersed_lags():
    """
    Test matrix of lags if list with interspersed lags.
    """
    forecaster = ForecasterRecursiveClassifier(LogisticRegression(), lags=[2, 3])
    results = forecaster._create_lags(y=np.arange(10))
    expected = (
        np.array([[1., 0.],
                  [2., 1.],
                  [3., 2.],
                  [4., 3.],
                  [5., 4.],
                  [6., 5.],
                  [7., 6.]]),
        np.array([3., 4., 5., 6., 7., 8., 9.])
    )

    np.testing.assert_array_almost_equal(results[0], expected[0])
    np.testing.assert_array_almost_equal(results[1], expected[1])


def test_create_lags_output_pandas_use_native_categoricals_False():
    """
    Test matrix of lags is created properly when X_as_pandas=True but not 
    using native pandas categoricals.
    """
    y = np.array([1, 2, 3, 1, 2, 3, 1, 2, 3, 1], dtype=int)
    forecaster = ForecasterRecursiveClassifier(LogisticRegression(), lags=3)
    results = forecaster._create_lags(
        y=y, X_as_pandas=True, train_index=pd.date_range('2020-01-03', periods=7, freq='D')
    )

    expected = (
        pd.DataFrame(
            data = np.array([
                       [3, 2, 1],
                       [1, 3, 2],
                       [2, 1, 3],
                       [3, 2, 1],
                       [1, 3, 2],
                       [2, 1, 3],
                       [3, 2, 1]]
                   ),
            columns = ["lag_1", "lag_2", "lag_3"],
            index = pd.date_range('2020-01-03', periods=7, freq='D')
        ),
        np.array([1, 2, 3, 1, 2, 3, 1], dtype=int)
    )

    pd.testing.assert_frame_equal(results[0], expected[0])
    np.testing.assert_array_almost_equal(results[1], expected[1])


def test_create_lags_output_pandas_use_native_categoricals_True():
    """
    Test matrix of lags is created properly when X_as_pandas=True  
    using native pandas categoricals.
    """
    y = np.array([1, 2, 3, 1, 2, 3, 1, 2, 3, 1], dtype=int)
    forecaster = ForecasterRecursiveClassifier(LGBMClassifier(verbose=-1), lags=3)
    results = forecaster._create_lags(
        y=y, X_as_pandas=True, train_index=pd.date_range('2020-01-03', periods=7, freq='D')
    )

    expected = (
        pd.DataFrame(
            data = np.array([
                       [3, 2, 1],
                       [1, 3, 2],
                       [2, 1, 3],
                       [3, 2, 1],
                       [1, 3, 2],
                       [2, 1, 3],
                       [3, 2, 1]]
                   ),
            columns = ["lag_1", "lag_2", "lag_3"],
            index = pd.date_range('2020-01-03', periods=7, freq='D')
        ),
        np.array([1, 2, 3, 1, 2, 3, 1], dtype=int)
    )

    for col in expected[0].columns:
        expected[0][col] = pd.Categorical(
                               values     = expected[0][col],
                               categories = [1, 2, 3],
                               ordered    = False
                           )

    pd.testing.assert_frame_equal(results[0], expected[0])
    np.testing.assert_array_almost_equal(results[1], expected[1])


def test_create_lags_when_window_size_window_features_greater_than_max_lag():
    """
    Test matrix of lags created properly when window_size of 
    window_features is greater than max lag.
    """
    forecaster = ForecasterRecursiveClassifier(
        LogisticRegression(), lags=3, window_features=rolling
    )
    results = forecaster._create_lags(y=np.arange(10))
    expected = (
        np.array([[5., 4., 3.],
                  [6., 5., 4.],
                  [7., 6., 5.],
                  [8., 7., 6.]]),
        np.array([6., 7., 8., 9.])
    )

    np.testing.assert_array_almost_equal(results[0], expected[0])
    np.testing.assert_array_almost_equal(results[1], expected[1])


def test_create_lags_output_lags_None():
    """
    Test matrix of lags when lags=None.
    """
    forecaster = ForecasterRecursiveClassifier(
        LogisticRegression(), lags=None, window_features=rolling
    )
    results = forecaster._create_lags(y=np.arange(10))
    expected = (
        None,
        np.array([6., 7., 8., 9.])
    )

    assert results[0] == expected[0]
    np.testing.assert_array_almost_equal(results[1], expected[1])



# Unit test check_predict_input
# ==============================================================================
import re
import pytest
import numpy as np
import pandas as pd
from sklearn.exceptions import NotFittedError
from skforecast.utils import check_predict_input
from skforecast.utils import check_exog
from skforecast.utils import check_extract_values_and_index
from skforecast.exceptions import MissingValuesWarning
from skforecast.exceptions import MissingExogWarning
from skforecast.exceptions import IgnoredArgumentWarning
from skforecast.exceptions import UnknownLevelWarning

if pd.__version__ < '2.2.0':
    freq = "M"
else:
    freq = "ME"


def test_check_predict_input_NotFittedError_when_fitted_is_False():
    """
    Test NotFittedError is raised when fitted is False.
    """
    err_msg = re.escape(
        "This Forecaster instance is not fitted yet. Call `fit` with "
        "appropriate arguments before using predict."
    )
    with pytest.raises(NotFittedError, match = err_msg):
        check_predict_input(
            forecaster_name  = 'ForecasterRecursive',
            steps            = 5,
            is_fitted        = False,
            exog_in_         = False,
            index_type_      = None,
            index_freq_      = None,
            window_size      = None,
            last_window      = None,
            last_window_exog = None,
            exog             = None,
            exog_names_in_   = None,
            interval         = None,
            alpha            = None,
            max_step         = None,
            levels           = None,
            series_names_in_ = None
        )
        
        
def test_check_predict_input_ValueError_when_steps_int_lower_than_1():
    """
    Test ValueError is raised when steps is a value lower than 1.
    """
    steps = -5

    err_msg = re.escape(
        f"`steps` must be an integer greater than or equal to 1. Got {steps}."
    )
    with pytest.raises(ValueError, match = err_msg):
        check_predict_input(
            forecaster_name  = 'ForecasterRecursive',
            steps            = steps,
            is_fitted        = True,
            exog_in_         = False,
            index_type_      = None,
            index_freq_      = None,
            window_size      = None,
            last_window      = None,
            last_window_exog = None,
            exog             = None,
            exog_names_in_   = None,
            interval         = None,
            alpha            = None,
            max_step         = None,
            levels           = None,
            series_names_in_ = None
        )
        
        
def test_check_predict_input_ValueError_when_steps_list_lower_than_1():
    """
    Test ValueError is raised when steps is a list with a value lower than 1. 
    (`ForecasterDirect` and `ForecasterDirectMultiVariate`).
    """
    steps = [0, 1, 2]

    err_msg = re.escape(
        f"The minimum value of `steps` must be equal to or greater than 1. "
        f"Got {min(steps)}."
    )
    with pytest.raises(ValueError, match = err_msg):
        check_predict_input(
            forecaster_name  = 'ForecasterDirect',
            steps            = steps,
            is_fitted        = True,
            exog_in_         = False,
            index_type_      = None,
            index_freq_      = None,
            window_size      = None,
            last_window      = None,
            last_window_exog = None,
            exog             = None,
            exog_names_in_   = None,
            interval         = None,
            alpha            = None,
            max_step         = None,
            levels           = None,
            series_names_in_ = None
        )


def test_check_predict_input_ValueError_when_last_step_greater_than_max_step():
    """
    Test ValueError is raised when max(steps) > max_step. (`ForecasterDirect` 
    and `ForecasterDirectMultiVariate`).
    """
    steps = list(np.arange(20) + 1)
    max_step = 10

    err_msg = re.escape(
        f"The maximum value of `steps` must be less than or equal to "
        f"the value of steps defined when initializing the forecaster. "
        f"Got {max(steps)}, but the maximum is {max_step}."
    )
    with pytest.raises(ValueError, match = err_msg):
        check_predict_input(
            forecaster_name  = 'ForecasterDirectMultiVariate',
            steps            = steps,
            is_fitted        = True,
            exog_in_         = False,
            index_type_      = None,
            index_freq_      = None,
            window_size      = None,
            last_window      = None,
            last_window_exog = None,
            exog             = None,
            exog_names_in_   = None,
            interval         = None,
            alpha            = None,
            max_step         = max_step,
            levels           = None,
            series_names_in_ = None
        )


def test_check_predict_input_TypeError_when_ForecasterRecursiveMultiSeries_and_level_not_str_list_or_None():
    """
    Test TypeError is raised when `levels` is not a str, a list or None.
    """
    levels = 5

    err_msg = re.escape(
        "`levels` must be a `list` of column names, a `str` of a column name or `None`."
    )
    with pytest.raises(TypeError, match = err_msg):
        check_predict_input(
            forecaster_name  = 'ForecasterRecursiveMultiSeries',
            steps            = 5,
            is_fitted        = True,
            exog_in_         = False,
            index_type_      = None,
            index_freq_      = None,
            window_size      = None,
            last_window      = None,
            last_window_exog = None,
            exog             = None,
            exog_names_in_   = None,
            interval         = None,
            alpha            = None,
            max_step         = None,
            levels           = levels,
            series_names_in_ = ['1', '2']
        )


@pytest.mark.parametrize("levels     , series_names_in_", 
                         [('1'       , ['2', '3']), 
                          (['1']     , ['2', '3']), 
                          (['1', '2'], ['2', '3'])])
def test_check_predict_input_UnknownLevelWarning_when_ForecasterRecursiveMultiSeries_and_level_not_in_series_names_in__onehot(levels, series_names_in_):
    """
    Test UnknownLevelWarning is raised when `levels` is not in `self.series_names_in_` in a 
    ForecasterRecursiveMultiSeries.
    """
    last_window = pd.DataFrame(
        {'1': np.arange(10), '2': np.arange(10), '3': np.arange(10)}, 
        index = pd.date_range(start='01/01/2018', periods=10, freq=freq)
    )

    warn_msg = re.escape(
        "`levels` {'1'} were not included in training. The resulting "
        "one-hot encoded columns for this feature will be all zeros."
    )
    with pytest.warns(UnknownLevelWarning, match = warn_msg):
        check_predict_input(
            forecaster_name   = 'ForecasterRecursiveMultiSeries',
            steps             = 5,
            is_fitted         = True,
            exog_in_          = False,
            index_type_       = pd.DatetimeIndex,
            index_freq_       = freq,
            window_size       = 5,
            last_window       = last_window,
            last_window_exog  = None,
            exog              = None,
            exog_names_in_    = None,
            interval          = [2.5, 97.5],
            alpha             = None,
            max_step          = None,
            levels            = levels,
            series_names_in_  = series_names_in_,
            levels_forecaster = None,
            encoding          = 'onehot'
        )


@pytest.mark.parametrize("encoding", ['ordinal', 'ordinal_category'])
@pytest.mark.parametrize("levels     , series_names_in_", 
                         [('1'       , ['2', '3']), 
                          (['1']     , ['2', '3']), 
                          (['1', '2'], ['2', '3'])])
def test_check_predict_input_UnknownLevelWarning_when_ForecasterRecursiveMultiSeries_and_level_not_in_series_names_in_(encoding, levels, series_names_in_):
    """
    Test UnknownLevelWarning is raised when `levels` is not in `self.series_names_in_` in a 
    ForecasterRecursiveMultiSeries.
    """
    last_window = pd.DataFrame(
        {'1': np.arange(10), '2': np.arange(10), '3': np.arange(10)}, 
        index = pd.date_range(start='01/01/2018', periods=10, freq=freq)
    )

    warn_msg = re.escape(
        "`levels` {'1'} were not included in training. "
        "Unknown levels are encoded as NaN, which may cause the "
        "prediction to fail if the estimator does not accept NaN values."
    )
    with pytest.warns(UnknownLevelWarning, match = warn_msg):
        check_predict_input(
            forecaster_name   = 'ForecasterRecursiveMultiSeries',
            steps             = 5,
            is_fitted         = True,
            exog_in_          = False,
            index_type_       = pd.DatetimeIndex,
            index_freq_       = freq,
            window_size       = 5,
            last_window       = last_window,
            last_window_exog  = None,
            exog              = None,
            exog_names_in_    = None,
            interval          = None,
            alpha             = None,
            max_step          = None,
            levels            = levels,
            series_names_in_  = series_names_in_,
            levels_forecaster = None,
            encoding          = encoding
        )


@pytest.mark.parametrize("levels     , levels_forecaster", 
                         [('1'       , '2'), 
                          (['1']     , ['2', '3']), 
                          (['1', '2'], ['2', '3'])])
def test_check_predict_input_ValueError_when_ForecasterRnn_and_level_not_in_levels_forecaster(levels, levels_forecaster):
    """
    Test ValueError is raised when `levels` is not in `self.levels` 
    (levels_forecaster) in a ForecasterRnn.
    """
    err_msg = re.escape(
        f"`levels` names must be included in the series used during fit "
        f"({levels_forecaster}). Got {levels}."
    )
    with pytest.raises(ValueError, match = err_msg):
        check_predict_input(
            forecaster_name   = 'ForecasterRnn',
            steps             = 5,
            is_fitted         = True,
            exog_in_          = False,
            index_type_       = None,
            index_freq_       = None,
            window_size       = None,
            last_window       = None,
            last_window_exog  = None,
            exog              = None,
            exog_names_in_    = None,
            interval          = None,
            alpha             = None,
            max_step          = None,
            levels            = levels,
            series_names_in_  = None,
            levels_forecaster = levels_forecaster,
        )


def test_check_predict_input_ValueError_when_exog_is_none_and_exog_in_is_true():
    """
    """
    err_msg = re.escape(
        "Forecaster trained with exogenous variable/s. "
        "Same variable/s must be provided when predicting."
    )   
    with pytest.raises(ValueError, match = err_msg):
        check_predict_input(
            forecaster_name  = 'ForecasterRecursive',
            steps            = 5,
            is_fitted        = True,
            exog_in_         = True,
            index_type_      = None,
            index_freq_      = None,
            window_size      = None,
            last_window      = None,
            last_window_exog = None,
            exog             = None,
            exog_names_in_   = None,
            interval         = None,
            alpha            = None,
            max_step         = None,
            levels           = None,
            series_names_in_ = None
        )


def test_check_predict_input_ValueError_when_exog_is_not_none_and_exog_in_is_false():
    """
    """
    err_msg = re.escape(
        "Forecaster trained without exogenous variable/s. "
        "`exog` must be `None` when predicting."
    )   
    with pytest.raises(ValueError, match = err_msg):
        check_predict_input(
            forecaster_name  = 'ForecasterRecursive',
            steps            = 5,
            is_fitted        = True,
            exog_in_         = False,
            index_type_      = None,
            index_freq_      = None,
            window_size      = None,
            last_window      = None,
            last_window_exog = None,
            exog             = pd.Series(np.arange(10)),
            exog_names_in_   = None,
            interval         = None,
            alpha            = None,
            max_step         = None,
            levels           = None,
            series_names_in_ = None
        )


def test_check_predict_input_ValueError_when_last_window_not_stored_during_training_single_series():
    """
    """
    last_window = None
    err_msg = re.escape(
        "`last_window` was not stored during training. If you don't want "
        "to retrain the Forecaster, provide `last_window` as argument."
    )
    with pytest.raises(ValueError, match = err_msg):
        check_predict_input(
            forecaster_name  = 'ForecasterRecursive',
            steps            = 10,
            is_fitted        = True,
            exog_in_         = True,
            index_type_      = pd.RangeIndex,
            index_freq_      = None,
            window_size      = 5,
            last_window      = last_window,
            last_window_exog = None,
            exog             = pd.Series(np.arange(10)),
            exog_names_in_   = None,
            interval         = None,
            alpha            = None,
            max_step         = None,
            levels           = None,
            series_names_in_ = None
        )


@pytest.mark.parametrize("forecaster_name", 
                         ['ForecasterRecursiveMultiSeries',
                          'ForecasterDirectMultiVariate'], 
                         ids=lambda ft: f'forecaster_name: {ft}')
def test_check_predict_input_TypeError_when_last_window_is_not_pandas_DataFrame(forecaster_name):
    """
    `ForecasterRecursiveMultiSeries` and `ForecasterDirectMultiVariate`.
    """
    last_window = np.arange(5)

    err_msg = re.escape(
        f"`last_window` must be a pandas DataFrame. Got {type(last_window)}."
    )
    with pytest.raises(TypeError, match = err_msg):
        check_predict_input(
            forecaster_name  = forecaster_name,
            steps            = 10,
            is_fitted        = True,
            exog_in_         = True,
            index_type_      = pd.RangeIndex,
            index_freq_      = None,
            window_size      = 5,
            last_window      = last_window,
            last_window_exog = None,
            exog             = pd.Series(np.arange(10)),
            exog_names_in_   = None,
            interval         = None,
            alpha            = None,
            max_step         = None,
            levels           = '1',
            series_names_in_ = ['1', '2']
        )


@pytest.mark.parametrize("levels     , last_window", 
                         [('1'       , pd.DataFrame({'3': [1, 2, 3], '4': [1, 2, 3]})), 
                          (['1']     , pd.DataFrame({'3': [1, 2, 3], '4': [1, 2, 3]})), 
                          (['1', '2'], pd.DataFrame({'3': [1, 2, 3], '4': [1, 2, 3]}))], 
                         ids = lambda values: f'levels: {values}')
def test_check_predict_input_ValueError_when_levels_not_in_last_window_ForecasterRecursiveMultiSeries(levels, last_window):
    """
    Check ValueError is raised when levels are no the same as last_window column names.
    """
    last_window_cols = last_window.columns.to_list()
    missing_levels = set(levels) - set(last_window_cols)
    err_msg = re.escape(
        f"`last_window` must contain a column(s) named as the level(s) to be predicted. "
        f"The following `levels` are missing in `last_window`: {missing_levels}\n"
        f"Ensure that `last_window` contains all the necessary columns "
        f"corresponding to the `levels` being predicted.\n"
        f"    Argument `levels`     : {levels}\n"
        f"    `last_window` columns : {last_window_cols}\n"
        f"Example: If `levels = ['series_1', 'series_2']`, make sure "
        f"`last_window` includes columns named 'series_1' and 'series_2'."
    )
    with pytest.raises(ValueError, match = err_msg):
        check_predict_input(
            forecaster_name  = 'ForecasterRecursiveMultiSeries',
            steps            = 10,
            is_fitted        = True,
            exog_in_         = True,
            index_type_      = pd.RangeIndex,
            index_freq_      = None,
            window_size      = 2,
            last_window      = last_window,
            last_window_exog = None,
            exog             = pd.Series(np.arange(10)),
            exog_names_in_   = None,
            interval         = None,
            alpha            = None,
            max_step         = None,
            levels           = levels,
            series_names_in_ = ['1', '2']
        )


def test_check_predict_input_ValueError_when_series_names_in__not_last_window_ForecasterDirectMultiVariate():
    """
    Check ValueError is raised when column names of series using during fit do not
    match with last_window column names.
    """
    last_window = pd.DataFrame({'l1': [1, 2, 3], '4': [1, 2, 3]})
    series_names_in_ = ['l1', 'l2']

    err_msg = re.escape(
        "`last_window` columns must be the same as the `series` "
        "column names used to create the X_train matrix.\n"
        "    `last_window` columns    : ['l1', '4']\n"
        "    `series` columns X train : ['l1', 'l2']"
    )
    with pytest.raises(ValueError, match = err_msg):
        check_predict_input(
            forecaster_name  = 'ForecasterDirectMultiVariate',
            steps            = 10,
            is_fitted        = True,
            exog_in_         = True,
            index_type_      = pd.RangeIndex,
            index_freq_      = None,
            window_size      = 2,
            last_window      = last_window,
            last_window_exog = None,
            exog             = pd.Series(np.arange(10)),
            exog_names_in_   = None,
            interval         = None,
            alpha            = None,
            max_step         = None,
            levels           = None,
            series_names_in_ = series_names_in_
        )


def test_check_predict_input_TypeError_when_last_window_is_not_pandas_series():
    """
    """
    last_window = np.arange(5)
    err_msg = re.escape(
        f"`last_window` must be a pandas Series or DataFrame. Got {type(last_window)}."
    )
    with pytest.raises(TypeError, match = err_msg):
        check_predict_input(
            forecaster_name  = 'ForecasterRecursive',
            steps            = 10,
            is_fitted        = True,
            exog_in_         = True,
            index_type_      = pd.RangeIndex,
            index_freq_      = None,
            window_size      = 5,
            last_window      = last_window,
            last_window_exog = None,
            exog             = pd.Series(np.arange(10)),
            exog_names_in_   = None,
            interval         = None,
            alpha            = None,
            max_step         = None,
            levels           = None,
            series_names_in_ = None
        )


def test_check_predict_input_ValueError_when_length_last_window_is_lower_than_window_size():
    """
    """
    window_size = 10

    err_msg = re.escape(
        f"`last_window` must have as many values as needed to "
        f"generate the predictors. For this forecaster it is {window_size}."
    )
    with pytest.raises(ValueError, match = err_msg):
        check_predict_input(
            forecaster_name  = 'ForecasterRecursive',
            steps            = 10,
            is_fitted        = True,
            exog_in_         = True,
            index_type_      = pd.RangeIndex,
            index_freq_      = None,
            window_size      = window_size,
            last_window      = pd.Series(np.arange(5)),
            last_window_exog = None,
            exog             = pd.Series(np.arange(10)),
            exog_names_in_   = None,
            interval         = None,
            alpha            = None,
            max_step         = None,
            levels           = None,
            series_names_in_ = None
        )


def test_check_predict_input_MissingValuesWarning_when_last_window_has_missing_values():
    """
    """
    warn_msg = re.escape(
        "`last_window` has missing values. Most of machine learning models do "
        "not allow missing values. Prediction method may fail."
    )
    with pytest.warns(MissingValuesWarning, match = warn_msg):
        check_predict_input(
            forecaster_name  = 'ForecasterRecursive',
            steps            = 10,
            is_fitted        = True,
            exog_in_         = False,
            index_type_      = pd.RangeIndex,
            index_freq_      = None,
            window_size      = 5,
            last_window      = pd.Series([1, 2, 3, 4, 5, np.nan]),
            last_window_exog = None,
            exog             = None,
            exog_names_in_   = None,
            interval         = None,
            alpha            = None,
            max_step         = None,
            levels           = None,
            series_names_in_ = None
        )


def test_check_predict_input_TypeError_when_last_window_index_is_not_of_index_type():
    """
    """
    last_window = pd.Series(np.arange(10))
    index_type_ = pd.DatetimeIndex
    _, last_window_index = check_extract_values_and_index(
        data=last_window, data_label='`last_window`', ignore_freq=False, return_values=False
    )

    err_msg = re.escape(
        f"Expected index of type {index_type_} for `last_window`. "
        f"Got {type(last_window_index)}."
    )
    with pytest.raises(TypeError, match = err_msg):
        check_predict_input(
            forecaster_name  = 'ForecasterRecursive',
            steps            = 10,
            is_fitted        = True,
            exog_in_         = False,
            index_type_      = index_type_,
            index_freq_      = None,
            window_size      = 5,
            last_window      = last_window,
            last_window_exog = None,
            exog             = None,
            exog_names_in_   = None,
            interval         = None,
            alpha            = None,
            max_step         = None,
            levels           = None,
            series_names_in_ = None
        )


def test_check_predict_input_TypeError_when_last_window_index_frequency_is_not_index_freq():
    """
    """
    last_window = pd.Series(np.arange(10), index=pd.date_range(start='1/1/2018', periods=10, freq='D'))
    index_freq_ = 'Y'
    _, last_window_index = check_extract_values_and_index(
        data=last_window, data_label='`last_window`', ignore_freq=False, return_values=False
    )

    err_msg = re.escape(
        f"Expected frequency of type {index_freq_} for `last_window`. "
        f"Got {last_window_index.freq}."
    )
    with pytest.raises(TypeError, match = err_msg):
        check_predict_input(
            forecaster_name  = 'ForecasterRecursive',
            steps            = 10,
            is_fitted        = True,
            exog_in_         = False,
            index_type_      = pd.DatetimeIndex,
            index_freq_      = index_freq_,
            window_size      = 5,
            last_window      = last_window,
            last_window_exog = None,
            exog             = None,
            exog_names_in_   = None,
            interval         = None,
            alpha            = None,
            max_step         = None,
            levels           = None,
            series_names_in_ = None
        )


def test_check_predict_input_TypeError_when_exog_is_not_pandas_series_or_dataframe_multiseries():
    """
    """
    err_msg = re.escape(
        f"`exog` must be a pandas Series, DataFrame or dict. Got {type(np.arange(10))}."
    )
    with pytest.raises(TypeError, match = err_msg):
        check_predict_input(
            forecaster_name  = 'ForecasterRecursiveMultiSeries',
            steps            = 10,
            is_fitted        = True,
            exog_in_         = True,
            index_type_      = pd.DatetimeIndex,
            index_freq_      = freq,
            window_size      = 5,
            last_window      = pd.DataFrame(np.arange(10), columns=['l1'], index=pd.date_range(start='1/1/2018', periods=10, freq=freq)),
            last_window_exog = None,
            exog             = np.arange(10),
            exog_names_in_   = None,
            interval         = None,
            alpha            = None,
            max_step         = None,
            levels           = ['l1'],
            series_names_in_ = ['l1', 'l2']
        )


def test_check_predict_input_TypeError_when_exog_is_not_pandas_series_or_dataframe():
    """
    """
    err_msg = re.escape(
        f"`exog` must be a pandas Series or DataFrame. Got {type(np.arange(10))}."
    )
    with pytest.raises(TypeError, match = err_msg):
        check_predict_input(
            forecaster_name  = 'ForecasterRecursive',
            steps            = 10,
            is_fitted        = True,
            exog_in_         = True,
            index_type_      = pd.DatetimeIndex,
            index_freq_      = freq,
            window_size      = 5,
            last_window      = pd.DataFrame(np.arange(10), columns=['l1'], index=pd.date_range(start='1/1/2018', periods=10, freq=freq)),
            last_window_exog = None,
            exog             = np.arange(10),
            exog_names_in_   = None,
            interval         = None,
            alpha            = None,
            max_step         = None,
            levels           = ['l1'],
            series_names_in_ = ['l1', 'l2']
        )


def test_check_predict_input_TypeError_when_exog_dict_and_not_pandas_series_or_DataFrame():
    """
    """
    err_msg = re.escape(
        f"`exog` for series 'l1' must be a pandas Series or DataFrame. Got {type(np.arange(10))}"
    )
    with pytest.raises(TypeError, match = err_msg):
        check_predict_input(
            forecaster_name  = 'ForecasterRecursiveMultiSeries',
            steps            = 10,
            is_fitted        = True,
            exog_in_         = True,
            index_type_      = pd.DatetimeIndex,
            index_freq_      = freq,
            window_size      = 5,
            last_window      = pd.DataFrame(np.arange(10), columns=['l1'], index=pd.date_range(start='1/1/2018', periods=10, freq=freq)),
            last_window_exog = None,
            exog             = {'l1': np.arange(10)},
            exog_names_in_   = None,
            interval         = None,
            alpha            = None,
            max_step         = None,
            levels           = ['l1'],
            series_names_in_ = ['l1', 'l2']
        )


def test_check_predict_input_MissingExogWarning_when_exog_dict_and_no_key_for_some_levels():
    """
    """
    warn_msg = re.escape(
        "`exog` does not contain keys for levels {'l2'}. "
        "Missing levels are filled with NaN. Most of machine learning "
        "models do not allow missing values. Prediction method may fail."
    )
    with pytest.warns(MissingExogWarning, match = warn_msg):
        check_predict_input(
            forecaster_name  = 'ForecasterRecursiveMultiSeries',
            steps            = 10,
            is_fitted        = True,
            exog_in_         = True,
            index_type_      = pd.DatetimeIndex,
            index_freq_      = freq,
            window_size      = 5,
            last_window      = pd.DataFrame({'l1': np.arange(10), 'l2': np.arange(10)}, index=pd.date_range(start='1/1/2018', periods=10, freq=freq)),
            last_window_exog = None,
            exog             = {'l1': pd.DataFrame(np.arange(10), columns=['exog_1'], index=pd.date_range(start='11/1/2018', periods=10, freq=freq))},
            exog_names_in_   = ['exog_1'],
            interval         = None,
            alpha            = None,
            max_step         = None,
            levels           = ['l1', 'l2'],
            series_names_in_ = ['l1', 'l2']
        )


def test_check_predict_input_MissingValuesWarning_when_exog_has_missing_values():
    """
    """
    warn_msg = re.escape(
        "`exog` has missing values. Most of machine learning models do "
        "not allow missing values. Prediction method may fail."
    )
    with pytest.warns(MissingValuesWarning, match = warn_msg):
        check_predict_input(
            forecaster_name  = 'ForecasterRecursive',
            steps            = 3,
            is_fitted        = True,
            exog_in_         = True,
            index_type_      = pd.DatetimeIndex,
            index_freq_      = freq,
            window_size      = 2,
            last_window      = pd.Series(np.arange(10), index=pd.date_range(start='1/1/2018', periods=10, freq=freq)),
            last_window_exog = None,
            exog             = pd.Series([1, 2, 3, np.nan], index=pd.date_range(start='11/1/2018', periods=4, freq=freq), name='exog1'),
            exog_names_in_   = ['exog1'],
            interval         = None,
            alpha            = None,
            max_step         = None,
            levels           = None,
            series_names_in_ = None
        )


@pytest.mark.parametrize("steps", [10, [1, 2, 3, 4, 5, 6], [2, 6]], 
                         ids=lambda steps: f'steps: {steps}')
def test_check_predict_input_MissingValuesWarning_when_len_exog_is_less_than_steps_MultiSeries(steps):
    """
    """
    last_step = max(steps) if isinstance(steps, list) else steps
    warn_msg = re.escape(
        f"`exog` doesn't have as many values as steps "
        f"predicted, {last_step}. Missing values are filled "
        f"with NaN. Most of machine learning models do not "
        f"allow missing values. Prediction method may fail."
    )
    with pytest.warns(MissingValuesWarning, match = warn_msg):
        check_predict_input(
            forecaster_name  = 'ForecasterRecursiveMultiSeries',
            steps            = steps,
            is_fitted        = True,
            exog_in_         = True,
            index_type_      = pd.DatetimeIndex,
            index_freq_      = freq,
            window_size      = 5,
            last_window      = pd.Series(np.arange(10), name='l1', index=pd.date_range(start='1/1/2018', periods=10, freq=freq)).to_frame(),
            last_window_exog = None,
            exog             = pd.Series(np.arange(5), name='exog1', index=pd.date_range(start='11/1/2018', periods=5, freq=freq)),
            exog_names_in_   = ['exog1'],
            interval         = None,
            alpha            = None,
            max_step         = None,
            levels           = ['l1'],
            series_names_in_ = ['l1', 'l2']
        )
            

@pytest.mark.parametrize("steps", [10, [1, 2, 3, 4, 5, 6], [2, 6]], 
                         ids=lambda steps: f'steps: {steps}')
def test_check_predict_input_ValueError_when_len_exog_is_less_than_steps(steps):
    """
    """
    last_step = max(steps) if isinstance(steps, list) else steps
    err_msg = re.escape(
        f"`exog` must have at least as many values as "
        f"steps predicted, {last_step}."
    )
    with pytest.raises(ValueError, match = err_msg):
        check_predict_input(
            forecaster_name  = 'ForecasterRecursive',
            steps            = steps,
            is_fitted        = True,
            exog_in_         = True,
            index_type_      = pd.DatetimeIndex,
            index_freq_      = freq,
            window_size      = 5,
            last_window      = pd.Series(np.arange(10), index=pd.date_range(start='1/1/2018', periods=10, freq=freq)),
            last_window_exog = None,
            exog             = pd.Series(np.arange(5)),
            exog_names_in_   = None,
            interval         = None,
            alpha            = None,
            max_step         = None,
            levels           = None,
            series_names_in_ = None
        )


def test_check_predict_input_MissingExogWarning_when_exog_is_DataFrame_without_columns_in_exog_names_in__MultiSeries():
    """
    Raise MissingExogWarning when there are missing columns in `exog` when 
    Forecaster multi series.
    """
    exog = pd.DataFrame(np.arange(10).reshape(5, 2), columns=['col1', 'col2'])
    exog.index = pd.date_range(start='1/1/2018', periods=5, freq=freq)
    exog_names_in_ = ['col1', 'col3']

    warn_msg = re.escape(
        "{'col3'} not present in `exog`. All values will be NaN."
    )
    with pytest.warns(MissingExogWarning, match = warn_msg):
        check_predict_input(
            forecaster_name  = 'ForecasterRecursiveMultiSeries',
            steps            = 2,
            is_fitted        = True,
            exog_in_         = True,
            index_type_      = pd.DatetimeIndex,
            index_freq_      = freq,
            window_size      = 5,
            last_window      = pd.DataFrame(np.arange(10), columns=['l1'], index=pd.date_range(start='1/1/2018', periods=10, freq=freq)),
            last_window_exog = None,
            exog             = exog,
            exog_names_in_   = exog_names_in_,
            interval         = None,
            alpha            = None,
            max_step         = None,
            levels           = ['l1'],
            series_names_in_ = ['l1', 'l2']
        )


def test_check_predict_input_ValueError_when_exog_is_DataFrame_without_columns_in_exog_names_in_():
    """
    Raise ValueError when there are missing columns in `exog`.
    """
    exog = pd.DataFrame(np.arange(10).reshape(5, 2), columns=['col1', 'col2'])
    exog_names_in_ = ['col1', 'col3']

    err_msg = re.escape(
        f"Missing columns in `exog`. Expected {exog_names_in_}. "
        f"Got {exog.columns.to_list()}."
    )
    with pytest.raises(ValueError, match = err_msg):
        check_predict_input(
            forecaster_name  = 'ForecasterRecursive',
            steps            = 2,
            is_fitted        = True,
            exog_in_         = True,
            index_type_      = pd.DatetimeIndex,
            index_freq_      = freq,
            window_size      = 5,
            last_window      = pd.Series(np.arange(10), index=pd.date_range(start='1/1/2018', periods=10, freq=freq)),
            last_window_exog = None,
            exog             = exog,
            exog_names_in_   = exog_names_in_,
            interval         = None,
            alpha            = None,
            max_step         = None,
            levels           = None,
            series_names_in_ = None
        )


def test_check_predict_input_ValueError_when_exog_is_Series_with_no_name():
    """
    Raise ValueError when `exog` is a pandas Series with no name.
    """
    exog = pd.Series(np.arange(10))
    exog_names_in_ = ['exog1']

    err_msg = re.escape(
        "When `exog` is a pandas Series, it must have a name. Got None."
    )
    with pytest.raises(ValueError, match = err_msg):
        check_predict_input(
            forecaster_name  = 'ForecasterRecursive',
            steps            = 2,
            is_fitted        = True,
            exog_in_         = True,
            index_type_      = pd.DatetimeIndex,
            index_freq_      = freq,
            window_size      = 5,
            last_window      = pd.Series(np.arange(10), index=pd.date_range(start='1/1/2018', periods=10, freq=freq)),
            last_window_exog = None,
            exog             = exog,
            exog_names_in_   = exog_names_in_,
            interval         = None,
            alpha            = None,
            max_step         = None,
            levels           = None,
            series_names_in_ = None
        )


def test_check_predict_input_IgnoredArgumentWarning_when_exog_is_Series_with_name_not_in_exog_names_in__multiseries():
    """
    Raise IgnoredArgumentWarning when `exog` is a pandas Series and its name is not in 
    `exog_names_in_` when Forecaster multi series.
    """
    exog = pd.Series(np.arange(10), name='exog2')
    exog.index = pd.date_range(start='11/01/2018', periods=10, freq=freq)
    exog_names_in_ = ['exog1']

    last_window = pd.Series(
        np.arange(10), 
        index = pd.date_range(start='01/01/2018', periods=10, freq=freq), 
        name  = 'l1'
    ).to_frame()

    warn_msg = re.escape(
        "'exog2' was not observed during training. `exog` is ignored. "
        "Exogenous variables must be one of: ['exog1']."
    )
    with pytest.warns(IgnoredArgumentWarning, match = warn_msg):
        check_predict_input(
            forecaster_name  = 'ForecasterRecursiveMultiSeries',
            steps            = 2,
            is_fitted        = True,
            exog_in_         = True,
            index_type_      = pd.DatetimeIndex,
            index_freq_      = freq,
            window_size      = 5,
            last_window      = last_window,
            last_window_exog = None,
            exog             = exog,
            exog_names_in_   = exog_names_in_,
            interval         = None,
            alpha            = None,
            max_step         = None,
            levels           = ['l1'],
            series_names_in_ = ['l1', 'l2']
        )


def test_check_predict_input_ValueError_when_exog_is_Series_with_name_not_in_exog_names_in_():
    """
    Raise ValueError when `exog` is a pandas Series and its name is not in 
    `exog_names_in_`.
    """
    exog = pd.Series(np.arange(10), name='exog2')
    exog_names_in_ = ['exog1']

    err_msg = re.escape(
        "'exog2' was not observed during training. "
        "Exogenous variables must be: ['exog1']."
    )
    with pytest.raises(ValueError, match = err_msg):
        check_predict_input(
            forecaster_name  = 'ForecasterRecursive',
            steps            = 2,
            is_fitted        = True,
            exog_in_         = True,
            index_type_      = pd.DatetimeIndex,
            index_freq_      = freq,
            window_size      = 5,
            last_window      = pd.Series(np.arange(10), index=pd.date_range(start='1/1/2018', periods=10, freq=freq)),
            last_window_exog = None,
            exog             = exog,
            exog_names_in_   = exog_names_in_,
            interval         = None,
            alpha            = None,
            max_step         = None,
            levels           = None,
            series_names_in_ = None
        )


def test_check_predict_input_TypeError_when_exog_index_is_not_of_index_type_multiseries():
    """
    """
    exog = pd.Series(np.arange(10), name='exog1')
    exog.index = pd.RangeIndex(start=0, stop=10, step=1)
    index_type_ = pd.DatetimeIndex

    last_window = pd.Series(
        np.arange(10), 
        index = pd.date_range(start='01/01/2018', periods=10, freq=freq), 
        name  = 'l1'
    ).to_frame()

    err_msg = re.escape(
        f"Expected index of type {index_type_} for `exog`. "
        f"Got {type(exog.index)}."
    )
    with pytest.raises(TypeError, match = err_msg):
        check_predict_input(
            forecaster_name  = 'ForecasterRecursiveMultiSeries',
            steps            = 10,
            is_fitted        = True,
            exog_in_         = True,
            index_type_      = index_type_,
            index_freq_      = freq,
            window_size      = 5,
            last_window      = last_window,
            last_window_exog = None,
            exog             = exog,
            exog_names_in_   = ['exog1'],
            interval         = None,
            alpha            = None,
            max_step         = None,
            levels           = ['l1'],
            series_names_in_ = ['l1', 'l2']
        )


def test_check_predict_input_TypeError_when_exog_index_is_not_of_index_type():
    """
    """
    exog = pd.Series(np.arange(10), name='exog1')
    index_type_ = pd.DatetimeIndex
    check_exog(exog = exog)
    _, exog_index = check_extract_values_and_index(
        data=exog, data_label='`exog`', ignore_freq=True, return_values=False
    )

    err_msg = re.escape(
        f"Expected index of type {index_type_} for `exog`. "
        f"Got {type(exog_index)}."
    )
    with pytest.raises(TypeError, match = err_msg):
        check_predict_input(
            forecaster_name  = 'ForecasterRecursive',
            steps            = 10,
            is_fitted        = True,
            exog_in_         = True,
            index_type_      = index_type_,
            index_freq_      = freq,
            window_size      = 5,
            last_window      = pd.Series(np.arange(10), index=pd.date_range(start='1/1/2018', periods=10, freq=freq)),
            last_window_exog = None,
            exog             = exog,
            exog_names_in_   = ['exog1'],
            interval         = None,
            alpha            = None,
            max_step         = None,
            levels           = None,
            series_names_in_ = None
        )


def test_check_predict_input_ValueError_when_exog_index_does_not_follow_last_window_index_DatetimeIndex():
    """
    Raise exception if `exog` index does not start at the end of `last_window` index when DatetimeIndex.
    """
    exog_datetime = pd.Series(data=np.random.rand(10), name='exog1')
    exog_datetime.index = pd.date_range(start='2022-03-01', periods=10, freq=freq)
    lw_datetime = pd.Series(data=np.random.rand(10))
    lw_datetime.index = pd.date_range(start='2022-01-01', periods=10, freq=freq)

    expected_index = '2022-11-30 00:00:00'

    err_msg = re.escape(
        f"To make predictions `exog` must start one step "
        f"ahead of `last_window`.\n"
        f"    `last_window` ends at : {lw_datetime.index[-1]}.\n"
        f"    `exog` starts at : {exog_datetime.index[0]}.\n"
        f"    Expected index : {expected_index}."
    )
    with pytest.raises(ValueError, match = err_msg):
        check_predict_input(
            forecaster_name  = 'ForecasterRecursive',
            steps            = 10,
            is_fitted        = True,
            exog_in_         = True,
            index_type_      = pd.DatetimeIndex,
            index_freq_      = freq,
            window_size      = 5,
            last_window      = lw_datetime,
            last_window_exog = None,
            exog             = exog_datetime,
            exog_names_in_   = ['exog1'],
            interval         = None,
            alpha            = None,
            max_step         = None,
            levels           = None,
            series_names_in_ = None
        )


def test_check_predict_input_ValueError_when_exog_index_does_not_follow_last_window_index_RangeIndex():
    """
    Raise ValueError if `exog` index does not start at the end of `last_window` index when RangeIndex.
    """
    exog_datetime = pd.Series(data=np.random.rand(10), name='exog1')
    exog_datetime.index = pd.RangeIndex(start=11, stop=21)
    lw_datetime = pd.Series(data=np.random.rand(10))
    lw_datetime.index = pd.RangeIndex(start=0, stop=10)

    expected_index = 10

    err_msg = re.escape(
        f"To make predictions `exog` must start one step "
        f"ahead of `last_window`.\n"
        f"    `last_window` ends at : {lw_datetime.index[-1]}.\n"
        f"    `exog` starts at : {exog_datetime.index[0]}.\n"
        f"    Expected index : {expected_index}."
    )
    with pytest.raises(ValueError, match = err_msg):
        check_predict_input(
            forecaster_name  = 'ForecasterRecursive',
            steps            = 10,
            is_fitted        = True,
            exog_in_         = True,
            index_type_      = pd.RangeIndex,
            index_freq_      = freq,
            window_size      = 5,
            last_window      = lw_datetime,
            last_window_exog = None,
            exog             = exog_datetime,
            exog_names_in_   = ['exog1'],
            interval         = None,
            alpha            = None,
            max_step         = None,
            levels           = None,
            series_names_in_ = None
        )


def test_check_predict_input_ValueError_when_ForecasterStats_last_window_exog_is_not_None_and_exog_in_is_false():
    """
    Check ValueError is raised when last_window_exog is not None, but exog_in_     
    is False.
    """
    err_msg = re.escape(
        "Forecaster trained without exogenous variable/s. "
        "`last_window_exog` must be `None` when predicting."
    )   
    with pytest.raises(ValueError, match = err_msg):
        check_predict_input(
            forecaster_name  = 'ForecasterStats',
            steps            = 3,
            is_fitted        = True,
            exog_in_         = False,
            index_type_      = pd.RangeIndex,
            index_freq_      = None,
            window_size      = 3,
            last_window      = pd.Series(np.arange(5)),
            last_window_exog = pd.Series(np.arange(5)),
            exog             = None,
            exog_names_in_   = None,
            interval         = None,
            alpha            = None,
            max_step         = None,
            levels           = None,
            series_names_in_ = None
        )


def test_check_predict_input_TypeError_when_ForecasterStats_last_window_exog_is_not_pandas_Series_or_DataFrame():
    """
    """
    last_window_exog = np.arange(5)
    err_msg = re.escape(
        f"`last_window_exog` must be a pandas Series or a "
        f"pandas DataFrame. Got {type(last_window_exog)}."
    )
    with pytest.raises(TypeError, match = err_msg):
        check_predict_input(
            forecaster_name  = 'ForecasterStats',
            steps            = 10,
            is_fitted        = True,
            exog_in_         = True,
            index_type_      = pd.RangeIndex,
            index_freq_      = None,
            window_size      = 3,
            last_window      = pd.Series(np.arange(5)),
            last_window_exog = last_window_exog,
            exog             = pd.Series(np.arange(10), index=pd.RangeIndex(start=5, stop=15), name='exog1'),
            exog_names_in_   = ['exog1'],
            interval         = None,
            alpha            = None,
            max_step         = None,
            levels           = None,
            series_names_in_ = None
        )


def test_check_predict_input_ValueError_when_ForecasterStats_length_last_window_exog_is_lower_than_window_size():
    """
    """
    window_size = 10

    err_msg = re.escape(
        f"`last_window_exog` must have as many values as needed to "
        f"generate the predictors. For this forecaster it is {window_size}."
    )
    with pytest.raises(ValueError, match = err_msg):
        check_predict_input(
            forecaster_name  = 'ForecasterStats',
            steps            = 10,
            is_fitted        = True,
            exog_in_         = True,
            index_type_      = pd.RangeIndex,
            index_freq_      = None,
            window_size      = window_size,
            last_window      = pd.Series(np.arange(10)),
            last_window_exog = pd.Series(np.arange(5)),
            exog             = pd.Series(np.arange(10), index=pd.RangeIndex(start=10, stop=20), name='exog1'),
            exog_names_in_   = ['exog1'],
            interval         = None,
            alpha            = None,
            max_step         = None,
            levels           = None,
            series_names_in_ = None
        )


def test_check_predict_input_MissingValuesWarning_when_ForecasterStats_last_window_exog_has_missing_values():
    """
    """
    warn_msg = re.escape(
        "`last_window_exog` has missing values. Most of machine learning models "
        "do not allow missing values. Prediction method may fail."
    )
    with pytest.warns(MissingValuesWarning, match = warn_msg):
        check_predict_input(
            forecaster_name  = 'ForecasterStats',
            steps            = 10,
            is_fitted        = True,
            exog_in_         = True,
            index_type_      = pd.RangeIndex,
            index_freq_      = None,
            window_size      = 5,
            last_window      = pd.Series(np.arange(10)),
            last_window_exog = pd.Series([1, 2, 3, 4, 5, np.nan], name='exog1'),
            exog             = pd.Series(np.arange(10), index=pd.RangeIndex(start=10, stop=20), name='exog1'),
            exog_names_in_   = ['exog1'],
            interval         = None,
            alpha            = None,
            max_step         = None,
            levels           = None,
            series_names_in_ = None
        )


def test_check_predict_input_TypeError_when_ForecasterStats_last_window_exog_index_is_not_of_index_type():
    """
    """
    last_window_exog = pd.Series(np.arange(10), name='exog1')
    index_type_ = pd.DatetimeIndex
    _, last_window_exog_index = check_extract_values_and_index(
        data=last_window_exog, data_label='`last_window_exog`', return_values=False
    )

    err_msg = re.escape(
        f"Expected index of type {index_type_} for `last_window_exog`. "
        f"Got {type(last_window_exog_index)}."
    )
    with pytest.raises(TypeError, match = err_msg):
        check_predict_input(
            forecaster_name  = 'ForecasterStats',
            steps            = 10,
            is_fitted        = True,
            exog_in_         = True,
            index_type_      = index_type_,
            index_freq_      = freq,
            window_size      = 5,
            last_window      = pd.Series(np.arange(10), index=pd.date_range(start='1/1/2018', periods=10, freq=freq)),
            last_window_exog = last_window_exog,
            exog             = pd.Series(np.arange(10), index=pd.date_range(start='30/11/2018', periods=10, freq=freq), name='exog1'),
            exog_names_in_   = ['exog1'],
            interval         = None,
            alpha            = None,
            max_step         = None,
            levels           = None,
            series_names_in_ = None
        )


def test_check_predict_input_TypeError_when_ForecasterStats_last_window_exog_index_frequency_is_not_index_freq():
    """
    """
    
    last_window_exog = pd.Series(np.arange(10), index=pd.date_range(start='2018', periods=10, freq='YE'))
    index_freq_ = freq
    _, last_window_exog_index = check_extract_values_and_index(
        data=last_window_exog, data_label='`last_window_exog`', return_values=False
    )

    err_msg = re.escape(
        f"Expected frequency of type {index_freq_} for `last_window_exog`. "
        f"Got {last_window_exog_index.freq}."
    )
    with pytest.raises(TypeError, match = err_msg):
        check_predict_input(
            forecaster_name  = 'ForecasterStats',
            steps            = 10,
            is_fitted        = True,
            exog_in_         = True,
            index_type_      = pd.DatetimeIndex,
            index_freq_      = index_freq_,
            window_size      = 5,
            last_window      = pd.Series(np.arange(10), index=pd.date_range(start='1/1/2018', periods=10, freq=freq)),
            last_window_exog = last_window_exog,
            exog             = pd.Series(np.arange(10), index=pd.date_range(start='30/11/2018', periods=10, freq=freq), name='exog1'),
            exog_names_in_   = ['exog1'],
            interval         = None,
            alpha            = None,
            max_step         = None,
            levels           = None,
            series_names_in_ = None
        )


def test_check_predict_input_ValueError_when_last_window_exog_is_DataFrame_without_columns_in_exog_names_in_():
    """
    Raise ValueError when there are missing columns in `last_window_exog`, ForecasterStats.
    """
    
    exog = pd.DataFrame(np.arange(10).reshape(5, 2), columns=['col1', 'col3'])
    exog.index = pd.date_range(start='6/1/2018', periods=5, freq=freq)
    exog_names_in_ = ['col1', 'col3']

    last_window_exog = pd.DataFrame(np.arange(10).reshape(5, 2), columns=['col1', 'col2'])
    last_window_exog.index = pd.date_range(start='1/1/2018', periods=5, freq=freq)

    err_msg = re.escape(
        f"Missing columns in `last_window_exog`. Expected {exog_names_in_}. "
        f"Got {last_window_exog.columns.to_list()}."
    )
    with pytest.raises(ValueError, match = err_msg):
        check_predict_input(
            forecaster_name  = 'ForecasterStats',
            steps            = 2,
            is_fitted        = True,
            exog_in_         = True,
            index_type_      = pd.DatetimeIndex,
            index_freq_      = freq,
            window_size      = 5,
            last_window      = pd.Series(np.arange(5), index=pd.date_range(start='1/1/2018', periods=5, freq=freq)),
            last_window_exog = last_window_exog,
            exog             = exog,
            exog_names_in_   = exog_names_in_,
            interval         = None,
            alpha            = None,
            max_step         = None,
            levels           = None,
            series_names_in_ = None
        )


# @pytest.mark.skipif(pd.__version__ > '2.2.0', reason="requires pandas < 2.2.0")
def test_check_predict_input_ValueError_when_last_window_exog_is_Series_with_no_name():
    """
    Raise ValueError when `last_window_exog` has no name, ForecasterStats.
    """
    
    exog = pd.Series(np.arange(5), name='exog1')
    exog.index = pd.date_range(start='6/1/2018', periods=5, freq=freq)
    exog_names_in_ = ['exog1']

    last_window_exog = pd.Series(np.arange(5))
    last_window_exog.index = pd.date_range(start='1/1/2018', periods=5, freq=freq)
    
    err_msg = re.escape(
        "When `last_window_exog` is a pandas Series, it must have a name. Got None."
    )
    with pytest.raises(ValueError, match = err_msg):
        check_predict_input(
            forecaster_name  = 'ForecasterStats',
            steps            = 2,
            is_fitted        = True,
            exog_in_         = True,
            index_type_      = pd.DatetimeIndex,
            index_freq_      = freq,
            window_size      = 5,
            last_window      = pd.Series(np.arange(5), index=pd.date_range(start='1/1/2018', periods=5, freq=freq)),
            last_window_exog = last_window_exog,
            exog             = exog,
            exog_names_in_   = exog_names_in_,
            interval         = None,
            alpha            = None,
            max_step         = None,
            levels           = None,
            series_names_in_ = None
        )


def test_check_predict_input_ValueError_when_last_window_exog_is_Series_without_name_in_exog_names_in_():
    """
    Raise ValueError when `last_window_exog` name is not in `exog_names_in_`, ForecasterStats.
    """
    exog = pd.Series(np.arange(5), name='exog1')
    exog.index = pd.date_range(start='6/1/2018', periods=5, freq=freq)
    exog_names_in_ = ['exog1']

    last_window_exog = pd.Series(np.arange(5), name='exog2')
    last_window_exog.index = pd.date_range(start='1/1/2018', periods=5, freq=freq)
    
    err_msg = re.escape(
        "'exog2' was not observed during training. "
        "Exogenous variables must be: ['exog1']."
    )
    with pytest.raises(ValueError, match = err_msg):
        check_predict_input(
            forecaster_name  = 'ForecasterStats',
            steps            = 2,
            is_fitted        = True,
            exog_in_         = True,
            index_type_      = pd.DatetimeIndex,
            index_freq_      = freq,
            window_size      = 5,
            last_window      = pd.Series(np.arange(5), index=pd.date_range(start='1/1/2018', periods=5, freq=freq)),
            last_window_exog = last_window_exog,
            exog             = exog,
            exog_names_in_   = exog_names_in_,
            interval         = None,
            alpha            = None,
            max_step         = None,
            levels           = None,
            series_names_in_ = None
        )
